/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (collect.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Mar 15 15:20:28 GMT 2000

  Updated:    Joerg Schulenburg 2004-2005

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: collect.C 

  PURPOSE: collect same configurations to reduce loops

  ---------------------------------------------------

*/

// #include <iostream.h>
#include <stdlib.h>
// #include<math.h>
#include <stdio.h>
// #include <string.h>

#include "../include/Constants.h"
#include "../include/Externals.h"
// #include "../include/FunctionPrototypes.h"
// #include "../include/nr.h"
// #include "../include/nrutil.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#ifndef DO_COLLECT
#define DO_COLLECT 1  /* set to 0 to switch off collect, default = 1 */
#endif

/* Joerg05:
 * who is calling  collect_term?
 *  62 times by obtain()+term2()..term62() in obtain.C 
 * parameters and its ranges?
 *  indices[0] - configuration (lattice animal), 1..NumberConfigurations
 *                                          new: 0 for GS Energy
 *  indices[1] - term 2..62, 999999=end (not used in solve!)
 *  indices[2..5] -  0,1..NumConfig ??  (0,0,0,0)=not_used(check it)
 * Outputfile fp1=*.input is used for solve()
 * Outputfile fp1=*_gs    is used for ...     (new Sep05) ???
 * maximum number of terms and clusters for sample models?
 *  MAX_NUMBER_TERMS_IN_CCM_EQNS to big?
 * ToDo: dynamic arrays
 */

#define MAX_out 1000

#if DO_COLLECT

// work: indices will be sorted (not used in the calling routine), Joerg Sep05
//       same or permutated indices[2..5] will be stored as sum only once
//       indices pointing to configurations 1..NumberConfigurations
//       permutated indices dont have the same probability! that means:
//       its difficult to simplify (no easy symmetry)
//       but it looks like that every term appears twice (ToDo: simplify!)
void collect_term(int indices[6],double dummy, FILE *fp1) 
{

  int i, j, k, config_index, term_index;
  static int n_out=0; /* count output terms, avoid GB of debugging output */
  static int n_in=0;  /* count input terms */
  static int counter=0;         /* buffer index */
  static int max_counter=0;     /* maximum buffer size */
  static int    descriptors[MAX_NUMBER_TERMS_IN_CCM_EQNS+1][6]; // 24 byte
  static double multipliers[MAX_NUMBER_TERMS_IN_CCM_EQNS+1];    //  8 byte
  int found;
  // it is very slow to search a big table for an entry sequentialy
  // so we introduce a hash table of indizes to that entries
  //  and a table of indices to that table to have fast sequential access 
  //  to spare filled hash table (Joerg Sep05, xxzLC-14 5x faster)
  // number of hashed entries is in counter
  static int hash[MAX_NUMBER_TERMS_IN_CCM_EQNS+1]; // hashed index to entries
  static int hidx[MAX_NUMBER_TERMS_IN_CCM_EQNS+1]; // index to hash entries
  static int hcol=0;  // counter for hash collisions, just as info
  static int hhit=0;  // counter for hash hits
  static int hnew=0;  // counter for new hash entries
  static int hval=-1;  // hash value

  n_in++;
  if (hval<0) { hval=0; // used as flag, initialize hash once!
    for (i=0;i<MAX_NUMBER_TERMS_IN_CCM_EQNS+1;i++) { hidx[i]=hash[i]=0; }
  }

  if (counter){  // something stored into buffer
      /* save collected terms and reset buffer for next term */

    if (    descriptors[counter][0] < indices[0]  /* next configuration number? */
//  Joerg Sep05: indices[1]=term not used in solve, so we can collect it too
//       || descriptors[counter][1] < indices[1]  /* next term? 2..62 */
       ||                       0 > indices[0]  /* flush buffer condition */
      /* Sept 05: old was 0>i[0] but collect_gse needs 0>=i[0] */
       ||    NumberConfigurations < indices[0]
       || counter >= MAX_NUMBER_TERMS_IN_CCM_EQNS) /* buffer full */
    {
        if (       MAX_NUMBER_TERMS_IN_CCM_EQNS == counter) {
           printf("MAX_NUMBER_TERMS_IN_CCM_EQNS is to small, flush buffer\n");
           printf("Overflow happens for cfg= %d  term= %d (%d %d %d %d)\n",
             indices[0],indices[1],
             indices[2],indices[3],indices[4],indices[5]);
           fflush(stdout);
        }
	// first save information 
	for(j=1;j<=counter;j++){
	  
	  fwrite(  descriptors[j] ,sizeof(   int),6,fp1);
	  fwrite(&(multipliers[j]),sizeof(double),1,fp1);
	  
	  // printf out information
	  
	  if (Control.output > 1 && myid == 0 && n_out < MAX_out) {
	    printf("OUT: Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d) : "
		   "mult=%6.3f\n",
		   descriptors[j][0],descriptors[j][1],
		   descriptors[j][2],descriptors[j][3],
                   descriptors[j][4],descriptors[j][5],
		   multipliers[j]);
            if (n_out+1==MAX_out)
              printf("excessive output is stopped after %d terms\n",MAX_out);
          }
          n_out++;
          hash[ hidx[j] ]=0; // reset hash entry
          hidx[j]=0; // not necessary
	  
	  // now overwrite the old stored information 
	  
	}
        for (i=1;i<=counter;i++) { hash[ hidx[i] ]=0; }
	counter=0;
      }
  }

  if (Control.output > 2 && myid == 0 && n_out < MAX_out)
    printf("IN:  Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)"
           " : mult=%6.3f counter=%d\n",
       indices[0],indices[1],
       indices[2],indices[3],indices[4],indices[5],dummy,counter);

  // the last config has lot of terms, try to reduce further by
  // reducing all possible permutations to one config by sorting (big first)

  for(;;) { // bubble sort indices
    j=1;
    for(i=2; i<5; i++)
      if (indices[i]<indices[i+1]) {  // exchange two indices
        j=0; k=indices[i+1]; indices[i+1]=indices[i]; indices[i]=k; 
      }
    if (j) break;
  }

#if 0 // old sequential search (slow for big m)
  // ToDo: this becomes slow, we could try a hash table here, Joerg
  //       then the reset is resetting hash table, having max_buffer prime
  j=0;
  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  for(j=1;j<=counter;j++) {
    if (   indices[0]==descriptors[j][0]
//      && indices[1]==descriptors[j][1]  // not used in solve (Joerg Sep05)
        && indices[2]==descriptors[j][2]
        && indices[3]==descriptors[j][3]
        && indices[4]==descriptors[j][4]
        && indices[5]==descriptors[j][5]
          ) { multipliers[j]+=dummy; break; }
  }
#else // new hash table search (fast)
  j=0;
  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  {
    hval=(((((                          indices[5])
              *(NumberConfigurations+1)+indices[4])
              *(NumberConfigurations+1)+indices[3])
              *(NumberConfigurations+1)+indices[2])
              *(NumberConfigurations+1)+indices[0]);
    // try 5 following entries
    for(i=0;;i++) {
      hval = abs(hval + indices[2]);  // avoid overflow to negativ values
      hval %= MAX_NUMBER_TERMS_IN_CCM_EQNS; // MAX should be prime
      j = hash[hval];
      if (j>counter || j<0)
         { printf("Bug86757 j=%d hash=%d\n",j,hval); exit(1); }
      if (j==0) {      // empty entry
             hash[hval]=counter+1;  // set hash entry
             hidx[counter+1]=hval;  // set sequential index to entry
             j=counter+1;           // store to table
             hnew++; break;
      } else {
        if ( indices[0]==descriptors[j][0] &&
             indices[2]==descriptors[j][2] &&
             indices[3]==descriptors[j][3] &&
             indices[4]==descriptors[j][4] &&
             indices[5]==descriptors[j][5]
           ) {
            multipliers[j]+=dummy;  // hashed == searched
            hhit++; break;
         } else if (i>=8) {
           hcol++; n_out++;    // collision, write through
           fwrite(indices,sizeof(int),6,fp1);
           fwrite(&dummy,sizeof(double),1,fp1);
           if (hcol<20) {
             printf("collision hash[%d]=%d\n", hval, hash[hval]);
             printf("IN:   Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     indices[0],indices[1],
                     indices[2],indices[3],
                     indices[4],indices[5]);
             printf("HASH: Config=%3d : Term = %2d : CCM = (%d,%d,%d,%d)\n",
                     descriptors[j][0],descriptors[j][1],
                     descriptors[j][2],descriptors[j][3],
                     descriptors[j][4],descriptors[j][5]);
           }
           j=0; break;
         }
      }
    }
  }
#endif
      
  if (indices[0]>=0 && indices[0]<=NumberConfigurations)
  if ( j>counter /* && indices[1] != 0 */ ) {
    counter++;  /* potential overflow! */
    if (counter>max_counter) max_counter=counter; /* just as info */
    descriptors[counter][0]=indices[0];
    descriptors[counter][1]=indices[1];
    descriptors[counter][2]=indices[2];
    descriptors[counter][3]=indices[3];
    descriptors[counter][4]=indices[4];
    descriptors[counter][5]=indices[5];
    multipliers[counter]=dummy;
  }

  if (indices[0]>NumberConfigurations || indices[0]<0) {
    /* flush buffer, called once per bond? */
    fwrite(indices,sizeof(int),6,fp1);
    fwrite(&dummy,sizeof(double),1,fp1);

    if (Control.output > 1 && myid == 0 ) {
      printf("OUT: Config=%3d : Term = %2d\n", indices[0], indices[1]);
    }
    if (Control.output > 0 && myid == 0 ) {
      /* output debugging infos */
      printf("collect_term myid= %d maxbuff= %d  terms in= %d  out= %d\n",
              myid, max_counter, n_in-1, n_out);
      max_counter=0; n_out=n_in=0;
      printf("collect_term myid= %d hash collisions= %d hits= %d new= %d \n",
              myid, hcol, hhit, hnew);
      hnew=hhit=hcol=0;   // reset number of collisions
    }
  }
   
}
#else /* dont collect */
void collect_term(int indices[6],double dummy, FILE *fp1) 
{
  static int n_out=0;

  fwrite(indices,sizeof(int),6,fp1);
  fwrite( &dummy,sizeof(double),1,fp1);
	  
  if (Control.output > 1 && myid == 0 && n_out < MAX_out) {
     printf("OUT: Config=%d : Term = %d : CCM = (%d,%d,%d,%d) : "
     "mult=%10.7f\n",
     indices[0],indices[1],
     indices[2],indices[3],
     indices[4],indices[5],
     dummy);
     if (n_out+1==MAX_out)
        printf("excessive output is stopped after %d terms\n", MAX_out);
  }
  if (indices[0]>NumberConfigurations || indices[0]<0) {
    if (Control.output > 0 && myid == 0 ) {
      /* output debugging infos */
      printf("collect_term myid= %d  terms out= %d\n", myid, n_out);
      n_out=0;
    }
  }
  n_out++;
}
#endif

