/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (define.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Sep 29 1999
 
  Purpose:    Determine and solve CCM LSUBm equations.

  Updated:    Joerg Schulenburg 2005

  ===================================================

  ---------------------------------------------------

  MODULE NAME: DEFINE.C 

  PURPOSE:     Defines the LSUBm clusters recursively.
               Not optimised. 

  ---------------------------------------------------

*/

//#include <iostream.h> /* gcc doesn like this, its obsolete */
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#include"../include/nr.h"
#include"../include/nrutil.h"
#include"../include/Constants.h"
#include"../include/Externals.h"
#include"../include/FunctionPrototypes.h"

#define MAX_out 100

void define(void)
{

  /* this module defines all of the cluster configurations using
     a recursive algorithm. 

     In essence this is a souped-up version of the algorithm used
     in the original ``define configurations'' C program. Here, for
     complex crystallogrphic lattices and higher spin.

  */

  int i,j,k,l,m,n,tempLatticeSymmetries;
  int cell;

  int NumberConnected=0;
  int NumberDisconnected=0;
  int restriction;

  char path[]="../Data/";
  char configurations[80];

  int **configuration;
  int descriptor[MAX_SPIN_LIST];
  int length,descriptorIndex;

  int **tempConfig;
  tempConfig=matrix(0,MAX_LSUBM+1,0,5);

  
  strcpy(configurations,Control.name);
  strcat(configurations,".configs");
  
  fp1=fopen(configurations,"w");  
  if (!fp1) {
    printf("Cannot Open %s...\nExiting!!\n\n",configurations);
    exit(1);
  }

  configuration=matrix(0,MAX_LSUBM+1,0,5);

  if (Control.output > 0 && myid == 0) {
    printf("\n============================================================\n");
    printf( "             DEFINE CLUSTERS: START ALGORITHM.\n");
    printf("============================================================\n\n");
  }

  /* !! NEW ALGORITHM FOR S>= 1/2.... 30-8-2000 */

  /* PART 1: DETERMINE THE LATTICE ANIMALS */

  printf("PART 1: DETERMINE THE LATTICE ANIMALS \n\n");

  NumberLatticeAnimals=0;
  NumberConfigurations=0;

  tempLatticeSymmetries=LatticeSymmetries.number;
  LatticeSymmetries.number=1;

  for(cell=1;cell<=UnitCell.number;cell++) {
    length=1;
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=cell;      
    recursiveClusteringFunction(configuration,
				length,
				Params.m,
				NumberLatticeAnimals);
    printf("define: UnitCell.number: %d\n", cell);
    printf("define: number of lattice animals: %4d\n", NumberLatticeAnimals);
  }


  if (Control.output > 2 && myid == 0) {
    /* PRINT OUT INFORMATION ON CLUSTERS */
    printf("PRINT OUT INFORMATION ON CLUSTERS:\n\n");
    for(i=1;i<=NumberLatticeAnimals;i++) {
      printf("Lattice animal number %4d:\n",i);
      printOutClusterFromDescriptor(latticeAnimals[i].descriptor);
    }
  }
   
  printf("\nDEFINE LATTICE ANIMALS: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 2: DETERMINE THE FUNDAMTENTAL SPIN CONFIGURATIONS */
  
  printf("\nPART 2: DEFINE FUNDAMENTAL SPIN CONFIGURATIONS.\n\n");

  LatticeSymmetries.number=tempLatticeSymmetries;
  findFundamentalSpinConfigurations();

  printf("\nDEFINE FUND. SPIN CONFIGS.: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS */
  
  printf("\nPART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS.\n\n"); 
  
  if (Control.output > 0 && myid == 0) {
    for(i=1;i<=NumberConfigurations && i<=MAX_out; i++) {
      printf("Cluster Number (%d): \n",i);
      printOutDescriptor(spinConfigurations[i].descriptor);
      printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
      printf("\n");
      if (i==MAX_out) printf("define: output of configs stopped\n");
    }
  }
  printf("\n");
  printf("Number of Connected Clusters    = %d (ToDo)\n", NumberConnected);
  printf("Number of Disconnected Clusters = %d (ToDo)\n", NumberDisconnected);
  printf("Total Number of Clusters        = %d\n", NumberConfigurations);
  printf("\n");

  fwrite(&NumberConfigurations,sizeof(int),1,fp1);
  for(i=1;i<=NumberConfigurations;i++) {
    fwrite(spinConfigurations[i].descriptor,
	   ((MAX_SPIN_LIST)*sizeof(int)), 1, fp1);
  }

  fclose(fp1);

   /* NEW!!: work out the maximum radius for all clusters 
     This is used to reduce the amount of sweeping for 
     F_k F_m s_k^z s_m^z -like terms. */

  radius=0;

  for(i=1;i<=NumberConfigurations;i++) {

    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     tempConfig);

    /* note that midPointX, midPointY, midPointZ indicate Unit cells */

    // note that integer rounding shouldn't affect this 
    // algorithm

    // now find the maximum `radius'

    for(j=1;j<=length;j++) {
      if (abs(tempConfig[j][1]) > radius) radius=abs(tempConfig[j][1]);
      if (abs(tempConfig[j][2]) > radius) radius=abs(tempConfig[j][2]);
      if (abs(tempConfig[j][3]) > radius) radius=abs(tempConfig[j][3]);
    }
  }

  radius+=1;

  printf("\nRadius = %d\n\n",radius);

  printf("DEFINE CLUSTERS: ALGORITHM SUCCESFULLY COMPLETED!\n\n");

}

// in:   configuration=matrix(0,MAX_LSUBM+1,0,5)={0,0,0,1} initially
//       length=1 initially
//       maxorder=Params.m=m(SUBn-m) distance?
// work: configuration=matrix(0,MAX_LSUBM+1,0,5)
// out:  latticeAnimals[].descriptor via updateDescriptors
//       NumberAnimals  (=NumberLatticeAnimals)
void recursiveClusteringFunction(int **configuration, 
				 int length, 
				 int maxorder, 
				 int &NumberAnimals) {

  /* A simple recursive function that determines all connected configs */

  int i,j,k,l, descriptorIndex, numberFlips, maxNumberFlips;
  int index, tmp[4];
  static int descriptor[MAX_SPIN_LIST];
  
  // first check whether we have an error using the `length' of cluster. 

  if (length > maxorder) {
    printf("Error in Recursion: length > maxorder!\nExiting...\n\n");
    exit(1); 
  } 

  // first check if the configuration has been seen before
  
  if (Control.output > 2 && myid == 0) {
    printf("Temporary Connected Cluster:\n"); // Temp = Recursive
    for(i=1;i<=length;i++)   // ( x,  y,  z, cell)
      printf("lattice site(%d): (%d, %d, %d, %d)\n",i,
	     configuration[i][1],
	     configuration[i][2],
	     configuration[i][3],
	     configuration[i][4]);	     
  }
  
  findUniqueDescriptorOfCluster(length, configuration, descriptor);
  descriptorIndex = findDescriptorIndex2(descriptor, NumberAnimals);
  
  if (Control.output > 2 && myid == 0)
    printf("Has it been previously found? = %d\n\n", descriptorIndex);
  
  if ((!descriptorIndex) && (length == maxorder)) {
    // store descriptor to latticeAnimals[].descriptor
    updateDescriptors2(descriptor,NumberAnimals);
    NumberAnimals++;
    if (Control.output > 1 && myid == 0) {
      printf("Lattice Animal Number (%d): \n",NumberAnimals);
      printOutDescriptor(descriptor);
      printOutClusterFromDescriptor(descriptor);
      printf("\n");
    }
  }
  
  if (length == maxorder) {

    // this is a `dummy case' which just makes sure that we don't
    // go over the maximum LSUBm approximation number adjacent
    // lattice sites for the lattice animals.

  } else {

    // length = 0...MAX_LSUBM
    if (length > MAX_LSUBM) {
       printf("Error overflow: length=%d > MAX_LSUBM\n", length);
       exit(1);
    }
    for(i=1;i<=length;i++) {
      index = configuration[i][4];  // cell
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.

	tmp[0]=configuration[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=configuration[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=configuration[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=LatticeNeighbours.cell[index][j];
	
        if (tmp[0] < 0) continue;
	
	// check that we have NOT repeated a given site.
	
	for(numberFlips=1,k=1;k<length+1;k++)
	  if (tmp[0] == configuration[k][1] &&
	      tmp[1] == configuration[k][2] &&
	      tmp[2] == configuration[k][3] &&
	      tmp[3] == configuration[k][4]) break;

        if (k<length+1) continue;

        // add new neighbouring site to the lattice animal
	configuration[length+1][1] = tmp[0];
	configuration[length+1][2] = tmp[1];
	configuration[length+1][3] = tmp[2];
	configuration[length+1][4] = tmp[3];
	
	recursiveClusteringFunction(configuration,length+1,maxorder,
				      NumberAnimals);
      }
    }
  }
}

// Joerg Sep05: simplify code
// compare descriptors
// return: 1 desc1>desc2, 0 desc1==desc2, -1 desc1<desc2
int cmpDescriptors(int *desc1, int *desc2) {
  int i;
  for(i=MAX_SPIN_LIST-1;i>0;i--) {
    if (desc2[i]<desc1[i]) { return  1; } else // greater
    if (desc2[i]>desc1[i]) { return -1; }      // smaller
  }
  return 0; // equal
}

// out: NumberConfigurations
// called by define() - DEFINE FUNDAMENTAL SPIN CONFIGURATIONS
void findFundamentalSpinConfigurations(void)
{

  int i,j,k,l;
  int max[MAX_LSUBM+5];  // BLOCK_LENGTH < MAX_LSUBM+5 ??? Joerg
  int max_num_spin_flips;
  int descriptor[MAX_SPIN_LIST+2];
  int temp_descriptor[MAX_SPIN_LIST+2];
  int index1,index2,index3;
  int temp1, temp2, temp3;
  int descriptorIndex;
  int length;
  int **configuration;
  int *spin;
  int **latticeAnimal;
  int tempLength;
  int flag, maxNumberFlips;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
//  static int configuration[MAX_LSUBM+1][5];
  spin=ivector(0,BLOCK_LENGTH*MAX_SPIN_LIST+2);
  latticeAnimal=imatrix(0,MAX_LSUBM+1,0,5);

  // for 1D spin-one this is simple.
  
  max_num_spin_flips = Params.s2 + 1; // 2*s+1
  
  // initialize array max[i] with (2s+1)^(i-1) 
  for(max[0]=1,i=1;i<=BLOCK_LENGTH;i++) {   // default: 10, (2s+1)^10 < 32bit?
    max[i] = max[i-1] * max_num_spin_flips;
    if (Control.output > 10 && myid == 0 && i<100) 
      printf("max[%d] = %d\n", i, max[i]);
    if (max[i] * 1.0 != max[i-1] * 1.0 * max_num_spin_flips) {
      printf("max[i] >= biggest INTEGER possible...\nExiting!!\n\n");
      exit(1);
    }
  }

  NumberConfigurations=0;
  
  // split Params.m into summands less_equal than BLOCK_LENGTH
  if (Params.m > (3*BLOCK_LENGTH)) {  // SUBn-m (m) n=#flips m=#dist
    printf("Define: Params.m > (3*BLOCK_LENGTH)...\nExiting!!\n\n");
    exit(1);
  }
  k=Params.m;
  temp3 = ((k>BLOCK_LENGTH)?BLOCK_LENGTH:k); k-=temp3;
  temp2 = ((k>BLOCK_LENGTH)?BLOCK_LENGTH:k); k-=temp2;
  temp1 = ((k>BLOCK_LENGTH)?BLOCK_LENGTH:k); k-=temp1;
  
  for(i=1;i<=NumberLatticeAnimals;i++) {
    
    // see bookKeeping.C
    length=findClusterFromDescriptor(latticeAnimals[i].descriptor,
				     latticeAnimal);

    if (length != Params.m) { // happens for xxzCubic-8-8 Joerg Sep05
      printf("Error findFundamentalSpinConfigurations:"
             " length=%d != Params.m=%d for LatticeAnimal=%d"
             "\nExiting!!\n\n", length, Params.m, i);
      printOutClusterFromDescriptor(latticeAnimals[i].descriptor);
      Control.output=5;
      length=findClusterFromDescriptor(latticeAnimals[i].descriptor,
                                     latticeAnimal);
      exit(1);
    }

    // slowest part of the program? Joerg
    for(index1=0;index1<max[temp1];index1++) {
      for(index2=0;index2<max[temp2];index2++) {
	for(index3=1;index3<max[temp3];index3++) {
	  
	  for(k=0;k<=MAX_SPIN_LIST;k++) 
	    temp_descriptor[k]=0;
	  
	  temp_descriptor[1]=index3;
	  temp_descriptor[2]=index2;
	  temp_descriptor[3]=index1;
	  
          // decode counter index[1..3] to code (0..2s,0..2s,0..2s)
	  findSpinState(temp_descriptor,spin);

          // spin[1..m]: 000 001 002 ... 00(2s) 010 011 012 ... 01(2s) ...

	  tempLength=0;
	  flag=0;

	  for(j=1;j<=Params.m;j++) {
	    maxNumberFlips= UnitCell.s2[latticeAnimal[j][4]];
	    if (spin[j] > maxNumberFlips) { flag=1; break; } // ignore
	    for(l=1;l<=spin[j];l++) { // copy this config 0..2s times
	      tempLength++;
              if (tempLength > MAX_LSUBM) continue; // Overflow Error below
              // ToDo: does it make sence? makes program slow
	      configuration[tempLength][1] = latticeAnimal[j][1];
	      configuration[tempLength][2] = latticeAnimal[j][2];
	      configuration[tempLength][3] = latticeAnimal[j][3];
	      configuration[tempLength][4] = latticeAnimal[j][4];
	    }
	  }

          if (tempLength > MAX_LSUBM) {
            printf("Error findFundamentalSpinConfigurations:"
                   " tempLength=%d > MAX_LSUBM\n", tempLength);
            exit(1);
          }
	  	  
	  if (tempLength == 0){
	    printf("DEFINE: tempLength == 0...\nExiting!!\n\n");
	    exit(1);
	  }

	  if (flag==0) {

            // ToDo Joerg: is it necessary to check all symmetrized
            //             configs here? Its slow! 
	    findUniqueDescriptorOfCluster(tempLength,
					  configuration,
					  descriptor);

	    descriptorIndex=findDescriptorIndex(descriptor,
						NumberConfigurations);
	    
	    if (Control.output > 2 && myid == 0) {
              printf("test spin configuration:\n");
	      printOutDescriptor(temp_descriptor);
	      printOutClusterFromDescriptor(temp_descriptor);
              if (cmpDescriptors(temp_descriptor, descriptor)!=0) { // Joerg
                printf("transformed into:\n");
 	        printOutDescriptor(descriptor);
	        printOutClusterFromDescriptor(descriptor);
              }
	      printf("Has it been previously found? = %d", descriptorIndex);
	      printf(" : Subset == %d",subsetClusters(descriptor));
              printf("\n\n");
	    }
	    
	    if ((!descriptorIndex) && subsetClusters(descriptor)) {
	      
	      updateDescriptors(descriptor,NumberConfigurations);
	      NumberConfigurations++;
	      
	      if (Control.output > 1 && myid == 0) {
                printf("add #config = %d:\n", NumberConfigurations);
		printOutDescriptor(descriptor);
		printOutClusterFromDescriptor(descriptor);
		printf("\n");
	      }
	    }
	  }
	}
      }
    }
  }

  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(latticeAnimal,0,MAX_LSUBM+1,0,5);
  free_ivector(spin,0,BLOCK_LENGTH*MAX_SPIN_LIST+2);
}

void obtainDisconnectedClusters(int &NumberDisconnected, 
				int NumberConnected)
{
}

void updateDescriptors(int *descriptor,
		       int numberElements)
{

  int i,j,k,l;
  int position;
  int tempLogicFlag;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=cmpDescriptors(descriptor,
             spinConfigurations[j].descriptor);
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) {
      for(i=1;i<MAX_SPIN_LIST;i++) {
	spinConfigurations[j+1].descriptor[i]=
	  spinConfigurations[j].descriptor[i];
      }
    }
  }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    spinConfigurations[position].descriptor[i]=descriptor[i];
  
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	printf("%d ", spinConfigurations[j].descriptor[i]);
      printf("\n");
    }
    printf("\n");
  }
}

// store descriptor to sorted latticeAnimals[].descriptor
void updateDescriptors2(int *descriptor,
			int numberElements)
{
  
  int i,j,k,l;
  int position;
  int tempLogicFlag;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=cmpDescriptors(descriptor,latticeAnimals[j].descriptor);
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) {
      for(i=1;i<MAX_SPIN_LIST;i++) {
	latticeAnimals[j+1].descriptor[i]=
	  latticeAnimals[j].descriptor[i];
      }
    }
  }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    latticeAnimals[position].descriptor[i]=descriptor[i];
  
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	printf("%d ", latticeAnimals[j].descriptor[i]);
      printf("\n");
    }
    printf("\n");
  }
}

int subsetClusters(int *descriptor)
{

  int i,j,k,l;
  int length;
  int **config;
  int flag;
  int temp;
  double x, y, z,stz;
  int int_x, int_y, int_z;

  config=imatrix(0,MAX_LSUBM+1,0,5);

  // this routine is somewhat lattice specific and is used to 
  // restrict the clusters used to within some well-defined
  // subset.

  // Hence, this subroutine is application-dependent and to be 
  // written by the user.

  // it returns a value of 0 if the cluster is acceptable
  // and a non-zero integer otherwise.

  length=findClusterFromDescriptor(descriptor,config);

  flag=1;

  // always check that we don't have more than n spin flips.

  if (length > Params.n) flag=0; 

  // restriction 1: NO RESTRICTION!

  // restriction 2: even number of spin flips.

  if (restrictionClusters==2) {
    if (flag==1) {
      if (length%2!=0)
	flag=0;
    }
  }
  
  // restriction 3: S_T^z=0

  if (restrictionClusters==3) {

    // USER DEFINED OPTION!!

    // ADVANCED: USER MUST ALTER CODE IN THIS CASE!!!

    // EXAMPLE: XXZ model or Heisenberg linear-chain, square or cubic
    // lattices arbitrary `s' though.

    // EXAMPLE: Ca V_4 O_9 Neel model state. Take out comment markers
    // on `x+=0.5;' and `y+=0.5;' and recompile

    if (flag==1) {
      stz=0.0;
      for(i=1;i<=length;i++) {
	findCartesiansFromGeneralisedPosition(x,y,z,config[i]);
	// for CAVO applications and SRCUBO now too!

	   x+=0.5000000;
	   y+=0.5000000;

	if (((int)(((fabs(x)+fabs(y)+fabs(z))+0.001)/2.0)) == 
	    ((int)(((fabs(x)+fabs(y)+fabs(z))+1.001)/2.0)))
	// For CAVO with striped model state 
	//	if (((int)(((fabs(y)+0.001)/2.0))) == ((int)((fabs(y)+1.001)/2.0)))
	  stz++;
	else
	  stz--;
      }
      if (fabs(stz)>0.1)
	flag=0;
    }
  }

  free_imatrix(config,0,MAX_LSUBM+1,0,5);
  return flag;

}

// called by findFundamentalSpinConfigurations
// recoded by Joerg Sep05
void findSpinState(int *descriptor, int *spinState) 
{
  int i, j, temp, base;

  base = Params.s2 + 1; // code base (2s+1) (was max[i], Joerg Sep05)

  for(i=1;i<=MAX_SPIN_LIST;i++) {
    temp=descriptor[i];
    for(j=1;j<=BLOCK_LENGTH;j++) {
      spinState[(i-1)*BLOCK_LENGTH+j] = temp % base; // code digit 0..2s
      temp /= base;   // shift code down one digit
    }
  }
}

