#define SLOW_MPI 0   /* old version, obsolete */
#define MAX_out 1000
/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (solve.C)
  
  Written by: Damian J. J. Farnell

  Date:      Wed Mar 15 15:20:28 GMT 2000

  Updated:   Joerg Schulenburg 2004

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: SOLVE.C 

  PURPOSE: Routine that solves saved CCM
           equations using newton-raphson
	   or direct iteration. Sublattice
           magnetisation determined and 
	   ground-state energy (+ derivatives).

  ---------------------------------------------------

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"
#include "../include/nr.h"
#include "../include/nrutil.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

char gs_name[100]="";
char input_name[200];
// char configurations[200];

/* -------------------------------------------------------------
 *  array of parameters depending on one variable parameter   
 *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2         
 */
double parameter_function(int bonds, double delta)
{
   static int once=0;
   // Joerg Sep05: changing parameterfunction gives you more flexibility
   if (once==0 && myid==0){ 
     printf("Parameter function can be defined in "__FILE__" L%d\n",__LINE__);
     once=1;
   }
    /* add your model specifique parameter here! */
    //  tdelta * H1 + H2 + H3 + ...
    if (bonds==1) return delta;

    //  cos(tdelta) * H1 + sin(tdelta) * H2 + H3 + ...
 // if (bonds==1) return cos(delta*M_PI);
 // if (bonds==2) return sin(delta*M_PI); 

    return 1.0;
}

void solve(void)
{

  int i, j, k, l, m, bonds,count;
  double *x, *dx;
  double *bra_x, *dBraX;
  char root_name[200],append[]=".dat";
  char output[200], output2[200];
  //  char path[]="../Data/";
  char v_j2[]="_v_j2";
  int  **configuration,length;
  int iterations,iterations2;
  double energy;  
  double energy2;
  double specHeatA, specHeatC;
  int status, stat;

  double stagMag;
  double Energy1;
  double Energy2;
  double dEnergy_diff;
  double ddEnergy_diff;
  double start, stop, tdelta;

  double subLatticeMag[MAX_UNIT_CELL+3]={0.0};

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  time_t theTime;

  //  solutionStrategy=1;
  //  momentum=0.7;

  if (PARALLEL_FLAG > 0 && solutionStrategy == 1){
    printf("Direct Iteration only for parallel processing...\n"
	   "Exiting!!\n\n");
    exit(0);
  }

  time(&theTime);
  srand((int) theTime);

   strcpy(input_name,    Control.name); strcat(input_name,    ".input");
   strcpy(output,        Control.name); strcat(output,        ".output");
// strcpy(configurations,Control.name); strcat(configurations,".configs");
   strcpy(gs_name,       Control.name); strcat(gs_name,       "_gs");

  printf(" solve: myid=%d input=%s output=%s gs=%s\n",
         myid,input_name,output,gs_name);

  fp=fopen(input_name,"r");   /* ToDo: set big buffer, Joerg Sep2005 */
  if (!fp) {
    printf("Cannot Open %s...\nExiting!!\n\n", input_name);
    exit(1);
  }

  //  fp1=fopen(configurations,"r");  
  // if (!fp) {
  //  printf("Cannot Open %s...\nExiting!!\n\n", configurations);
  //  exit(1);
  // }

  if (myid == 0) {
    fp2=fopen(output,"w");  
    if (!fp2) {
      printf("Cannot Open %s...\nExiting!!\n\n", output);
      exit(1);
    }
  }

  // ground-state energy equation.

  fp_gs=fopen(gs_name,"r");
  if (!fp_gs) {
    printf("Cannot Open %s...\nExiting!!\n\n", gs_name);
    exit(1);
  }
  
  Energy2=1.0;
  Energy1=2.0;
  // precondition is set in include/Constants.h to 2
  if (precondition)
    dx=   dvector(0,NumberConfigurations+5);
  x=      dvector(0,NumberConfigurations+5);
  if (precondition>1)
    dBraX=dvector(0,NumberConfigurations+5);
  bra_x=  dvector(0,NumberConfigurations+5);

  for(i=1;i<=NumberConfigurations;i++) { 

    if (precondition)   dx[i]=0.0; 
    // if (precondition>2)   // why ">2" ? mistake?
    if (precondition>1) dBraX[i]=0.0; 

    x[i]=    0.000000001; 
    bra_x[i]=0.000000001; 
  }
  x[0]=    1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  dx[0]=   1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  bra_x[0]=1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  dBraX[0]=1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */

  // x[2] = 0.064048;
  // x[3] = 0.000000;
  // x[4] = 0.000000;
  // x[5] = 0.064048;
  // x[6] = 0.064048;
 
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("============================================================\n\n");
  }

  count=0;
  if (increment > 0.0) { start= minimum;  stop= maximum; }
    else               { start=-maximum;  stop=-minimum; }

  for(tdelta=start; tdelta<=stop; tdelta=tdelta+fabs(increment)) {

    count++;

    for (int i=1; i<= Params.bonds; i++) {
      delta[i] = parameter_function(i,((increment<0.0)?-1:+1)*tdelta);
    }

    iterations=0;
    iterations2=0;

    
    if (solutionStrategy==1) {
      
     // (a) newton-raphson ket state + LU decomp. bra state.

      status=mnewt((int) MAX_NUM_ITERATIONS, NumberConfigurations, 
		   (double) TOLERANCE, (double) TOLERANCE, iterations, x);
      
      findBraState(x, bra_x);

    } else {
      
      // (b) direct iterations for both ket and bra states.

      status=directIteration((int) MAX_NUM_ITERATIONS,
             NumberConfigurations, (double) TOLERANCE, 
	     iterations, x, dx, count);

      status=directIterationBraState((int) MAX_NUM_ITERATIONS, 
	     NumberConfigurations, (double) TOLERANCE, 
	     iterations2, x, bra_x, dBraX, count); 
    }


    if (iterations >= MAX_NUM_ITERATIONS)
      break;
    
    energy=findEnergy(x);
    
    dEnergy_diff=(energy-Energy1)/(-increment);
    
    if (Control.output > 3 && myid == 0)
      printf("Computational First Derivative Eg wrt delta = %f\n",
         dEnergy_diff);
    
    ddEnergy_diff=-(energy-2*Energy1+Energy2)/(increment*increment); 
    
    // = -d^2 E_g/ d Delta^2 via computational evaluation only.
    
    Energy2=Energy1;
    Energy1=energy;
    
    // (c) the (sub)lattice magnetisation.
      
    stagMag=findStagMag(x,bra_x,subLatticeMag);
    
    // PRINT OUT INFORMATION.
    
    if (Control.output > -1 && myid == 0) {
      printf("Del. = %6.4f : Eg = %12.8f : M = %12.8f : x[1]=%12.8f\n"
	     "Heat = %12.8f : iterations = %d : iterations bra = %d\n",
	     tdelta,energy,stagMag,x[1],ddEnergy_diff,iterations,iterations2);
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  printf("Mag.[%d] = %12.8f ",k,subLatticeMag[k]);
	printf("\n");
      }
    }

    if (myid == 0) {
      fprintf(fp2,"%16.12f %16.12f %16.12f %16.12f %16.12f %d %d",
	      tdelta,energy,stagMag,ddEnergy_diff,x[1],iterations,iterations2);
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  fprintf(fp2,"%16.12f ",subLatticeMag[k]);
      }
      fprintf(fp2,"\n");
      fflush(fp2);
    }
  }
  if (Control.output > -1 && myid == 0) {
    printf("============================================================\n\n");
    printf("SOLVE CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
    printf("============================================================\n\n");
  }

  fclose(fp);    fp=NULL;
  fclose(fp_gs); fp_gs=NULL; 
  if (myid==0) fclose(fp2);

  //  stat=remove(gs_name);
  // stat=remove(input_name);  
  // stat=remove(configurations);

}

/* called by solve() and directIteration(only_printout)
 * new: x[0]=1 now (Sep05)
 * 
 */
double findEnergy(double *x) 
{

  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy;
  double sum;
  static int n_out=0;

  sum=0.0;
  double missing_term = 0;
  
  rewind(fp_gs);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      fread(indices,sizeof(int),   6,fp_gs);
      fread( &dummy,sizeof(double),1,fp_gs);
      
      if (indices[0]<0 || indices[0] > NumberConfigurations) break;
      
      // Joerg Sep05: why indices[4] is not used??? Bug ???
      //   according to gse_terms only indices[2..3] can be nonzero
      sum += dummy * x[indices[2]] * x[indices[3]] * x[indices[5]];
      missing_term +=
             dummy * x[indices[2]] * x[indices[3]]
                   * x[indices[4]] * x[indices[5]] /* correct? */
           - dummy * x[indices[2]] * x[indices[3]];

    }

    sum          *= delta[bonds];
    missing_term *= delta[bonds];

  }
  if (missing_term != 0.0)
   printf("ERROR: findEnergy: sum = %e Ediff = %e\n",sum, missing_term);
  return sum;
}
 
// in/ouut: oldX - startvector + iterationvector
int directIteration(int maximumNumIterations, 
		    int numConfigs, 
		    double tolerance,
		    int &iterations, 
		    double *oldX,
		    double *dx,
		    int count)
{  
  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4,multiplier;
  double *newX, *lhs, *rhs, error,*refX;
  double tempSum1, sum1, tempSum2, sum2;
  double *tempLHS,  *tempRHS;
  double *tempLHS2, *tempRHS2;

  if (precondition) 
    refX=dvector(0,numConfigs+5);

  newX=dvector(0,numConfigs+5);
  lhs= dvector(0,numConfigs+5);
  rhs= dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    if (precondition) 
      refX[i]=oldX[i];
    newX[i]=0.0;
    lhs[i]=0.0;
    rhs[i]=0.0;
  }
  
  /* now: oldX[0]=1.0, Joerg Sep05 */
  if (precondition && (count > 2)) {   // set startvector

    rewind(fp);
    
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      multiplier=delta[bonds];
      
      for(i=1;i<=1000000000;i++) { // max. 32GB
	
	fread(indices,sizeof(int),6,fp);
	fread(dummy,sizeof(double),1,fp);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations )
	  break;  // next bonds
	
	temp1=oldX[indices[2]]; /* oldX[0]=1 */
	temp2=oldX[indices[3]];
	temp3=oldX[indices[4]];
	temp4=oldX[indices[5]];
	
        /* three indices are 0, one is config_number, sorted? */
        /* Joerg: dependency on bonds not clear to me ??? just a startvec */
	if ((indices[2]==0 && indices[3]==0 && indices[4]==0 
	                        && indices[5]==indices[0] && bonds==2)
	 || (indices[2]==0 && indices[3]==0 && indices[5]==0 
                                && indices[4]==indices[0] && bonds==2)
	 || (indices[2]==0 && indices[4]==0 && indices[5]==0 
	  	                && indices[3]==indices[0] && bonds==2)
	 || (indices[3]==0 && indices[4]==0 && indices[5]==0 // this if sorted
		                && indices[2]==indices[0] && bonds==2)) {
	  lhs[indices[0]] -= dummy[0];
	} else {
	  
	  if (bonds==1)
	    rhs[indices[0]] += dummy[0]*temp1*temp2*temp3*temp4*increment;
	  
          dx[0]=0; /* simplify code, Joerg Sep05 */
          rhs[indices[0]] += ( dx[indices[2]]
                             + dx[indices[3]]
                             + dx[indices[4]]
                             + dx[indices[5]])
                           * dummy[0]*temp2*temp3*temp4*multiplier;
	}
      }
      printf("directIteration-1 : bonds = %d : nread = %d\n",bonds,i);
    }

#if (PARALLEL_FLAG> 0) 
    tempLHS2=dvector(0,numConfigs-1);
    tempRHS2=dvector(0,numConfigs-1);
    MPI_Allreduce(lhs+1,tempLHS2, numConfigs,
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1,tempRHS2,  numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS2[i-1];
      rhs[i]=tempRHS2[i-1];
    }
    free_dvector(tempLHS2,0,numConfigs-1);
    free_dvector(tempRHS2,0,numConfigs-1);    
#endif

    for(i=1;i<=numConfigs;i++) {
      if (fabs(lhs[i])>0.000000001)
	oldX[i] += rhs[i]/lhs[i];
      lhs[i]=0.0;
      rhs[i]=0.0;
    }
  }

  for(iterations=1; iterations<=maximumNumIterations; iterations++) {
    
    rewind(fp);

    /* calculate rhs and lhs from input-file */
    for(bonds=1; bonds<=Params.bonds; bonds++) {
      
      multiplier=delta[bonds];
      
      /* why loop over bonds? inner loop does not depend on bonds? */
      for(i=1;i<=1000000000;i++) {
	
	fread(indices,sizeof(int),6,fp);
	fread(dummy,sizeof(double),1,fp);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;
	
        /* now: oldX[0]=1.0, Joerg Sep05 */
        temp1=oldX[indices[2]];
        temp2=oldX[indices[3]];
        temp3=oldX[indices[4]];
        temp4=oldX[indices[5]];
    
        if ((indices[2]==0 && indices[3]==0 && indices[4]==0 
                 && indices[5]==indices[0])
        ||  (indices[2]==0 && indices[3]==0 && indices[5]==0 
                 && indices[4]==indices[0])
        ||  (indices[2]==0 && indices[4]==0 && indices[5]==0 
                 && indices[3]==indices[0])
        ||  (indices[3]==0 && indices[4]==0 && indices[5]==0 
                 && indices[2]==indices[0]))
          lhs[indices[0]]-=dummy[0]*multiplier;
        else {
          rhs[indices[0]]+=dummy[0]*temp1*temp2*temp3*temp4*multiplier;
        }
      }
    }

    error=0.0;

#if (PARALLEL_FLAG> 0) 
    error=0.0;
    /* MPI_Barrier(MPI_COMM_WORLD); dont needed? */
  { /* new faster code by JS */
    double *tmpSum1, *sum1;
    tmpSum1=dvector(0,numConfigs-1);
    sum1   =dvector(0,numConfigs-1);
    for(j=1;j<=numConfigs;j++) {   /* counted from 1 not from 0! */
      tmpSum1[j-1]=sum1[j-1]=0;
      if ((j%npe) == myid) { tmpSum1[j-1]=rhs[j]/lhs[j]; }
    }
    MPI_Allreduce(tmpSum1, sum1, numConfigs,
                  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(j=1;j<=numConfigs;j++) {
      newX[j]=momentum*oldX[j]+(1.0-momentum)*sum1[j-1];
      error+=fabs(oldX[j]-sum1[j-1]);
      oldX[j]=newX[j];
      lhs[j]=0.0;
      rhs[j]=0.0;
    }
    free_dvector(tmpSum1,0,numConfigs-1);
    free_dvector(sum1   ,0,numConfigs-1);    
  }
#else
    for(j=1;j<=numConfigs;j++) {
      if (fabs(lhs[j]) < tolerance) {
	printf("Error in direct iterations: no left-hand side...\n"
	       "Exiting!!\n\n");
	exit(1);
      }
      newX[j]=momentum*oldX[j]+(1.0-momentum)*rhs[j]/lhs[j];
      error+=fabs(oldX[j]-rhs[j]/lhs[j]);
      oldX[j]=newX[j];
      lhs[j]=0.0;
      rhs[j]=0.0;
    }
#endif
    
    /* calling findEnergy() here is a vast of time ??? (Joerg) */
    if (Control.output > 0 && myid == 0 && ((iterations % 20)==1)) 
      printf("Direct iteration ket state = %d : energy = %20.10f : "
	     "error = %20.10f\n",
	     iterations, findEnergy(oldX), error);

    if (error < tolerance)
      break;

  }

  if (precondition) {
    for(j=1;j<=numConfigs;j++) {
      dx[j]=oldX[j]-refX[j];
    }
  }

  if (iterations>=maximumNumIterations) {
    printf("directIteration Error: iterations>=maximumNumIterations\n");
    exit(0);
  }

  if (precondition) 
    free_dvector(refX,0,numConfigs+5);
  free_dvector(newX,0,numConfigs+5);
  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);

  return iterations;
  
}

// in: x[NumConf]
// out: fvec[NumConf], fjac[NumConf][NumConf]
// called by mnewt()=NewtonRawson in num_methods.C
void usrfun(int n, double *fvec, double **fjac, double *x, int iterations)
{  
  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4, multiplier;
  
  rewind(fp);

  for(i=1;i<=NumberConfigurations;i++) {    
    fvec[i]=0.0;
    for(k=1;k<=NumberConfigurations;k++) 
      fjac[i][k]=0.0;
  }  
 

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    multiplier=delta[bonds];
    
    for(i=1;i<=1000000000;i++) {
      
      fread(indices,sizeof(int),6,fp);
      fread(dummy,sizeof(double),1,fp);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations) 
	break;
      
      x[0]=1.0; /* obsolete??? Joerg Sep05 */
      temp1=x[indices[2]];
      temp2=x[indices[3]];
      temp3=x[indices[4]];
      temp4=x[indices[5]];
      
      fvec[indices[0]] += dummy[0]*temp1*temp2*temp3*temp4*multiplier;
      
      if (indices[2] > 0)
	fjac[indices[0]][indices[2]]+= dummy[0]*temp2*temp3*temp4*multiplier;
      
      if (indices[3] > 0)
	fjac[indices[0]][indices[3]]+= dummy[0]*temp1*temp3*temp4*multiplier;
      
      if (indices[4] > 0)
	fjac[indices[0]][indices[4]]+= dummy[0]*temp1*temp2*temp4*multiplier;
      
      if (indices[5] > 0)
	fjac[indices[0]][indices[5]]+= dummy[0]*temp1*temp2*temp3*multiplier;
    }
  }
  if (Control.output > 10 && myid == 0) {
    for(i=1;i<=NumberConfigurations;i++) {    
      for(k=1;k<=NumberConfigurations;k++) 
	printf("%4.2f ",fjac[i][k]);
      printf(" : %f \n",fvec[i]);
    }
  }
  
}

/*****************************************************
 *                                                   *
 * Obtain bra state (LU decomp.)                     *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * use LU decomposition to solve the equations:      *
 *                                                   *
 * F(I,J) (\tilde X(J)) = - e_o(I)                   *
 *                                                   *
 *****************************************************/

void findBraState(double *x, double *bra_x)
{

  /* Determines the bra-state correlation coefficients
     using LU decomposition.
     */

  int      i,j,k,flag;
  int      status;
  void     lubksb(double **a, int n, int *indx, double b[]);
  void     ludcmp(double **a, int n, int *indx, double *d);
  int      *ind;
  double   errx,errf,d,*fvec,**fjac, *p;
  double   **bra_eqn;
  int      length;
  int      iterations;

  for(i=1;i<=NumberConfigurations;i++) 
    bra_x[i]=0.0; 

  ind=ivector(0,NumberConfigurations+1);
  bra_eqn=dmatrix(0,NumberConfigurations+1,0,NumberConfigurations+1); 

  findBraStateRHS(bra_x,x);
  
  // find sum_J \tilde x_J (d(F_J)/d(x_I)) == F(I,J)
  
  findBraEquations(bra_eqn, x);
  
  // LU decomp Numerical Recipes to solve: F(I,J) (\tilde X(J)) = - e_o(I)
  
  ludcmp(bra_eqn,NumberConfigurations,ind,&d);
  lubksb(bra_eqn,NumberConfigurations,ind,bra_x);
  
  free_ivector(ind,0,NumberConfigurations+1);
  free_dmatrix(bra_eqn,0,NumberConfigurations+1,0,NumberConfigurations+1);

}
  

/*****************************************************
 *                                                   *
 * Obtain bra state (direct iteration)               *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * Use direct iteration to solve:                    *
 *                                                   *
 *    (\tilde X(I)) = { - e_o(I) - \sum_{J != I}     *
 *                F(I,J) (\tilde X(J))}/F(I,I)       *
 *                                                   *
 * for the \tilde X(I)'s.                            *
 *                                                   *
 *****************************************************/

int directIterationBraState(int maximumNumIterations, 
			    int numConfigs, 
			    double tolerance,
			    int &iterations, 
			    double *oldX,
			    double *oldBraX,
			    double *dBraX,
			    int     count)
{

  double error;
  
  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4,multiplier,sum;
  double tempSum1, sum1,tempSum2, sum2;
  double subLatticeMag[MAX_UNIT_CELL+3];

  double *newBraX, *lhs, *rhs, *tempLHS, *tempRHS;
  double *tempLHS2, *tempRHS2, *refBraX;
  
  if (precondition>1)
    refBraX=dvector(0,numConfigs+5);
  newBraX=dvector(0,numConfigs+5);
  lhs=dvector(0,numConfigs+5);
  rhs=dvector(0,numConfigs+5);

  for(i=1;i<=numConfigs;i++) {
    if (precondition>1)
      refBraX[i]=oldBraX[i];
    lhs[i]=0.0;
    rhs[i]=0.0;
  }

  if ((precondition>1) && (count > 2)) {
        
    rewind(fp);
    
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      multiplier=delta[bonds];
      
      for(i=1;i<=1000000000;i++) {
	
	fread(indices,sizeof(int),6,fp);
	fread(dummy,sizeof(double),1,fp);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;
	
        oldX[0]=1.0;
	temp1=oldX[indices[2]];
	temp2=oldX[indices[3]];
	temp3=oldX[indices[4]];
	temp4=oldX[indices[5]];
	
	if (indices[5]>0) {
	  if (indices[5]==indices[0] && indices[4]==0 && 
	      indices[3]==0 && indices[2] ==0 && bonds == 2)
	    lhs[indices[0]]+=dummy[0]*temp1*temp2*temp3;
	  else {
	    if (bonds==1)
	      rhs[indices[5]]-=dummy[0]*temp1*temp2*temp3*increment*
		oldBraX[indices[0]];
	    rhs[indices[5]]-=dummy[0]*temp1*temp2*temp3*multiplier*
	      dBraX[indices[0]];
	  }
	}
	
	if (indices[4]>0) {
	  if (indices[4]==indices[0]&& indices[5]==0 && 
	      indices[3]==0 && indices[2] ==0  && bonds == 2)
	    lhs[indices[0]]+=dummy[0]*temp1*temp2*temp4;
	  else {
	    if (bonds==1)
	      rhs[indices[4]]-=dummy[0]*temp1*temp2*temp4*increment*
		oldBraX[indices[0]];
	    rhs[indices[4]]-=dummy[0]*temp1*temp2*temp4*multiplier*
	      dBraX[indices[0]];
	  }
	}
	
	if (indices[3]>0) {
	  if (indices[3]==indices[0] && indices[5]==0 && 
	      indices[4]==0 && indices[2] ==0  && bonds == 2)
	    lhs[indices[0]]+=dummy[0]*temp1*temp3*temp4;
	  else {
	    if (bonds==1)
	      rhs[indices[3]]-=dummy[0]*temp1*temp3*temp4*increment*
		oldBraX[indices[0]];
	    rhs[indices[3]]-=dummy[0]*temp1*temp3*temp4*multiplier*
	      dBraX[indices[0]];
	  }
	}
	
	if (indices[2]>0) {
	  if (indices[2]==indices[0] && indices[5]==0 && 
	      indices[4]==0 && indices[3] ==0 && bonds == 2)
	    lhs[indices[0]]+=dummy[0]*temp2*temp3*temp4*multiplier;
	  else {
	    if (bonds==1)	      
	      rhs[indices[2]]-=dummy[0]*temp2*temp3*temp4*increment*
		oldBraX[indices[0]];
	    rhs[indices[2]]-=dummy[0]*temp2*temp3*temp4*multiplier*
	      dBraX[indices[0]];
	  }
	}
      }
    }
    
#if (PARALLEL_FLAG> 0) 
    MPI_Barrier(MPI_COMM_WORLD);
    tempLHS2=dvector(0,numConfigs-1);
    tempRHS2=dvector(0,numConfigs-1);
    MPI_Allreduce(lhs+1,tempLHS2, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1,tempRHS2, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS2[i-1];
      rhs[i]=tempRHS2[i-1];
    }
    free_dvector(tempLHS2,0,numConfigs-1);
    free_dvector(tempRHS2,0,numConfigs-1);    
#endif
    
    findBraStateRHS2(rhs,oldX);

    for(i=1;i<=numConfigs;i++) {
      if (fabs(lhs[i])>0.000000001)
	oldBraX[i]+=rhs[i]/lhs[i];
      lhs[i]=0.0;
      rhs[i]=0.0;
    }
  }

  for(iterations=1;iterations<=maximumNumIterations;iterations++) {
        
#if (PARALLEL_FLAG == 0) 
    findBraStateRHS(rhs,oldX);
#endif

    rewind(fp);

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      multiplier=delta[bonds];
      
      for(i=1;i<=1000000000;i++) {
	
	fread(indices,sizeof(int),6,fp);
	fread(dummy,sizeof(double),1,fp);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;
	
	if (indices[2]==0 && indices[3]==0 && indices[4]==0 && indices[5]==0) {
	  
	  //	  rhs[indices[0]]+=dummy[0]*multiplier;
	  
	} else {
	  
          oldX[0]=1.0;
	  temp1=oldX[indices[2]];
	  temp2=oldX[indices[3]];
	  temp3=oldX[indices[4]];
	  temp4=oldX[indices[5]];
	  
	  if (indices[5]>0) {
	    if (indices[5]==indices[0] && indices[4]==0 && 
		indices[3]==0 && indices[2] ==0)
	      lhs[indices[0]]+=dummy[0]*temp1*temp2*temp3*multiplier;
	    else 
	      rhs[indices[5]]-=dummy[0]*temp1*temp2*temp3*multiplier*
		oldBraX[indices[0]];
	  }
	  
	  if (indices[4]>0) {
	    if (indices[4]==indices[0]&& indices[5]==0 && 
	      	indices[3]==0 && indices[2] ==0)
	      lhs[indices[0]]+=dummy[0]*temp1*temp2*temp4*multiplier;
	    else 
	      rhs[indices[4]]-=dummy[0]*temp1*temp2*temp4*multiplier*
		oldBraX[indices[0]];
	  }

	  if (indices[3]>0) {
	    if (indices[3]==indices[0] && indices[5]==0 && 
	      	indices[4]==0 && indices[2] ==0)
	      lhs[indices[0]]+=dummy[0]*temp1*temp3*temp4*multiplier;
	    else
	      rhs[indices[3]]-=dummy[0]*temp1*temp3*temp4*multiplier*
		oldBraX[indices[0]];
	  }
	  
	  if (indices[2]>0) {
	    if (indices[2]==indices[0] && indices[5]==0 && 
	      	indices[4]==0 && indices[3] ==0)
	      lhs[indices[0]]+=dummy[0]*temp2*temp3*temp4*multiplier;
	    else 
	      rhs[indices[2]]-=dummy[0]*temp2*temp3*temp4*multiplier*
		oldBraX[indices[0]];
	  }
	}
      }
    }
      
#if (PARALLEL_FLAG> 0) 
    tempLHS=dvector(0,numConfigs-1);
    tempRHS=dvector(0,numConfigs-1);
    MPI_Allreduce(lhs+1, tempLHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1, tempRHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS[i-1];
      rhs[i]=tempRHS[i-1];
    }
    free_dvector(tempLHS,0,numConfigs-1);
    free_dvector(tempRHS,0,numConfigs-1);    

    findBraStateRHS(rhs,oldX);
#endif

    error=0.0;
    for(j=1;j<=numConfigs;j++) {

      if (fabs(lhs[j]) < tolerance) {
	printf("Error in direct iterations bra state: no left-hand side...\n"
	       "Exiting!!\n\n");
	exit(1);
      }
      newBraX[j]=momentum*oldBraX[j]+(1.0-momentum)*rhs[j]/lhs[j];
      error+=fabs(oldBraX[j]-rhs[j]/lhs[j]);

      oldBraX[j]=newBraX[j];
      lhs[j]=0.0;
      rhs[j]=0.0;
    }

    if (Control.output > 0 && myid == 0 && ((iterations % 20)==1)) 
      printf("Direct iteration bra state = %d : mag = %20.10f : "
	     "error = %20.10f\n",
	     iterations,findStagMag(oldX,oldBraX,subLatticeMag),error);

    if (error < tolerance)
      break;

  }

  if (precondition>1) {
    for(j=1;j<=numConfigs;j++) {
      dBraX[j]=oldBraX[j]-refBraX[j];
    }
  }
  
  if (iterations>=maximumNumIterations) {
    printf("directIterationBraState Error: i>=maximumNumIterations\n");
    exit(0);
  }

  if (precondition>1)
    free_dvector(refBraX,0,numConfigs+5);
  free_dvector(newBraX,0,numConfigs+5);
  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);

  return iterations;
}

void findBraStateRHS(double *rhs, double *x) 
{

  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4,multiplier,sum;

  sum=0.0;
  
  rewind(fp_gs);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    multiplier=delta[bonds];
    
    for(i=1;i<=1000000000;i++) {
      
      fread(indices,sizeof(int),6,fp_gs);
      fread(dummy,sizeof(double),1,fp_gs);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      if (!(indices[2]==0 && indices[3]==0 && indices[4]==0 && 
	    indices[5]==0)) {
	
          x[0]=1;
	  temp1=x[indices[2]];
	  temp2=x[indices[3]];
	  temp3=x[indices[4]];
	  temp4=x[indices[5]];
	
	if (indices[2] > 0)
	  rhs[indices[2]]-=
	    dummy[0]*temp2*temp3*temp4*multiplier;
	
	if (indices[3] > 0)
	  rhs[indices[3]]-=
	    dummy[0]*temp1*temp3*temp4*multiplier;
	
	if (indices[4] > 0)
	  rhs[indices[4]]-=
	    dummy[0]*temp1*temp2*temp4*multiplier;
	
	if (indices[5] > 0)
	  rhs[indices[5]]-=
	    dummy[0]*temp1*temp2*temp3*multiplier;
      }
    }
  }
}

void findBraStateRHS2(double *rhs, double *x) 
{

  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4,multiplier,sum;

  sum=0.0;
  
  rewind(fp_gs);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    multiplier=delta[bonds];
    
    for(i=1;i<=1000000000;i++) {
      
      fread(indices,sizeof(int),6,fp_gs);
      fread(dummy,sizeof(double),1,fp_gs);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      if (!(indices[2]==0 && indices[3]==0 && indices[4]==0 && 
	    indices[5]==0)) {
	
          x[0]=1.0;
	  temp1=x[indices[2]];
	  temp2=x[indices[3]];
	  temp3=x[indices[4]];
	  temp4=x[indices[5]];
	
	if (indices[2] > 0 && bonds==1)
	  rhs[indices[2]]-=
	    dummy[0]*temp2*temp3*temp4*increment;
	
	if (indices[3] > 0 && bonds==1)
	  rhs[indices[3]]-=
	    dummy[0]*temp1*temp3*temp4*increment;
	
	if (indices[4] > 0 && bonds==1)
	  rhs[indices[4]]-=
	    dummy[0]*temp1*temp2*temp4*increment;
	
	if (indices[5] > 0 && bonds==1)
	  rhs[indices[5]]-=
	    dummy[0]*temp1*temp2*temp3*increment;
      }
    }
  }
}

void findBraEquations(double **m, double *x) 
{

  /* FIND BRA EQUATIONS: finds the Jacobian in order to determine 
   the bra state equations.
   */

  // find: sum_J \tilde x_J (d(F_J)/d(x_I))
  
  int i,j,k,bonds,stat;
  int indices[7]={0};
  double dummy[1];
  double temp1, temp2, temp3, temp4, multiplier;
  
  rewind(fp);
  
  for(i=1;i<=NumberConfigurations;i++)    
    for(k=1;k<=NumberConfigurations;k++) 
      m[i][k]=0.0;

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    
    multiplier=delta[bonds];
 
    for(i=1;i<=1000000000;i++) {
      
      fread(indices,sizeof(int),6,fp);
      fread(dummy,sizeof(double),1,fp);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      if (!(indices[2]==0 && indices[3]==0 && indices[4]==0 && 
	    indices[5]==0)) {
	
        x[0]=1.0;
	  temp1=x[indices[2]];
	  temp2=x[indices[3]];
	  temp3=x[indices[4]];
	  temp4=x[indices[5]];
	
	if (indices[2] > 0)
	  m[indices[2]][indices[0]]+=
	    dummy[0]*temp2*temp3*temp4*multiplier;
	
	if (indices[3] > 0)
	  m[indices[3]][indices[0]]+=
	    dummy[0]*temp1*temp3*temp4*multiplier;
	
	if (indices[4] > 0)
	  m[indices[4]][indices[0]]+=
	    dummy[0]*temp1*temp2*temp4*multiplier;
	
	if (indices[5] > 0)
	  m[indices[5]][indices[0]]+=
	    dummy[0]*temp1*temp2*temp3*multiplier;
      }
    }
    
    if (i > 1000000000) {
      printf("Error i > 1000000000 in findBraEquations...\nExiting!!\n\n");
      exit(1);
    }
    
  }
}

/*****************************************************
 *                                                   *
 * Name: Staggered Magnetisation Routine.            *
 *                                                   *
 * Determines the lattice magnetisation once bra-    *
 * and ket-state equations have been solved.         *
 *****************************************************/

double findStagMag(double *x, double *bra_x, double *subLatticeMag)
{

  /* FIND STAG MAG: obtains the staggered magnetisation.
     */

  int      i,j,k,flag;
  double   stagMag;
  int      length;
  int      **configuration;
  double   subTerm[MAX_UNIT_CELL+3];

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  // first find the \sum_i < \tilde \Psi | s_i^z | \Psi > bit.

  stagMag=0.0; 
  
  for(i=1;i<=UnitCell.number;i++) {
    subLatticeMag[i] = 0.5*UnitCell.s2[i];
    stagMag         += 0.5*UnitCell.s2[i];    
  }
  
  stagMag=stagMag/((double) UnitCell.number);

  for(i=1;i<=NumberConfigurations;i++) {
    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     configuration);
    stagMag-=sublattice_term(i,length,configuration,bra_x,x,subTerm);
    
    for(j=1;j<=UnitCell.number;j++) 
      subLatticeMag[j]-=subTerm[j];
  }

  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return stagMag;
}

double sublattice_term(int config_index, 
		       int length, 
		       int **string,
		       double *bra_x,
		       double *x,
		       double *subTerm)
{
  
  /* 
  
     This routine finds contribution to the sublattice magnetisation
     from a given cluster, namely:
     
     F(m) . S(+)(m) 


  */
  
  int i,j,k,l,flag;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  static int tempConfiguration[MAX_LSUBM+1][5];
  
  int neighbour;
  int index;
  int new_index;
  double sum;

  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  // tempConfiguration=imatrix(0,MAX_LSUBM+1,0,5); // slow

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  sum+=length*bra_x[config_index]*x[config_index];    

  for(i=1;i<=UnitCell.number;i++)
    subTerm[i]=0.0;

   for(j=1;j<=LatticeSymmetries.number;j++) {
    
    // find symmetry related clusters
    
    symmetryGroup(j,length,configuration,tempConfiguration);
    
    for(i=1;i<=length;i++) {
      subTerm[tempConfiguration[i][4]] += bra_x[config_index]
                                         *    x[config_index]
                                         *    UnitCell.number;
    }
  } 

  for(i=1;i<=UnitCell.number;i++)
    subTerm[i]=subTerm[i]/LatticeSymmetries.number;
  
  /* for(i=1;i<=length;i++) {
    subTerm[configuration[i][4]]+=bra_x[config_index]*x[config_index]*
      UnitCell.number;
      }*/
  
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
 //  free_imatrix(tempConfiguration,0,MAX_LSUBM+1,0,5);
  return sum;
}

