/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (obtain.C)
  
  Copyright: Damian J. J. Farnell

  Date:      Mon Feb 21 2000

  Updated:   Joerg Schulenburg 2004-2005

  Purpose:   Determine CCM LSUBm equations.

  ===================================================

  ---------------------------------------------------

  MODULE NAME: OBTAIN.C 

  PURPOSE:     `Obtains' the CCM equations for LSUBm

  ---------------------------------------------------

  Code checked on Thu Aug 3 11:08:54 BST 2000, DJJF.

  ---------------------------------------------------

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "../include/nr.h"
#include "../include/nrutil.h"
#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"
#include "../include/terms.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

extern void collect_term(int indices[6],double dummy,FILE *fp);
   void collect_gse_term(int indices[6],double dummy,FILE *fp_gs);

void obtain(void)
{

  int    i,j,k,l,bonds;
  int    length;
  int    spin[MAX_LSUBM+2];
  int    **configuration;
  int    descriptor[MAX_SPIN_LIST];
  int    dummy;
  int    descriptorIndex;
  int    counter=0;
  double energy;
  double oldEnergy;
  double lhs, rhs;
  double diff;
  double momentum;
#define WBUF (1024*1024)    /* 1MB */
  static char wbuf[WBUF];    /* write buffer */ 
  char output[60];
  
  static int n_out=0;   /* number of outputs */
#define MAX_out 100
  
  char gs_name[100]="";

  momentum=0.3;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  printf("\n");

  // The data structures binary_max[], trimary_max[] ... are used 
  // to partition a given cluster into two (e.g. F(k) F(m)), three 
  // .... pieces in each `term' in H. 
  
  binary_max[0]=1;     dummy=2;   // 2^0 .. 2^30  # 30bit
  for(i=1;i<=30;i++) { binary_max[i]=dummy;   dummy=2*dummy; }

  trimary_max[0]=1;    dummy=3;   // 3^0 .. 3^10  # 16bit, i<=MAX_LSUBM
  for(i=1;i<=10;i++) { trimary_max[i]=dummy;  dummy=3*dummy; }

  quartary_max[0]=1;   dummy=4;   // 4^0 .. 4^10  # 20bit
  for(i=1;i<=10;i++) { quartary_max[i]=dummy; dummy=4*dummy; }

  strcpy(output,Control.name);      strcat(output,".input");
  strcpy(gs_name,Control.name);     strcat(gs_name,"_gs");

  fp=fopen(output,"w");

  // ground-state energy equation.

  fp_gs=fopen(gs_name,"w");  /* same content for all threads */

  if (!fp_gs) {
    printf("Problem opening files ground-state energies\nExiting!!\n\n");
    exit(1);
  }
      
  // find CCM equations.

  setvbuf(fp,wbuf,_IOFBF,WBUF);  /* set write buffer, minimize sys-time */
  if (myid==0){
    printf(" Write: %s %s (buffered)\n\n",gs_name,Control.name);
    printf("==========================================================\n\n");
    printf("OBTAIN CCM EQUATIONS: START ALGORITHM.\n\n");
    printf("==========================================================\n\n");
  }

  lhs=0.0;rhs=0.0;energy=0.0;
  /* delta=1.0;  Joerg Sep05: delta seems to be not used here ??? */

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    
    // find ground-state energy
    energy += /* delta * */ gse_term1(bonds, 1);
    energy += /* delta * */ gse_term2(bonds,27);
    energy += /* delta * */ gse_term3(bonds,30);
    energy += /* delta * */ gse_term4(bonds, 8);
    energy += /* delta * */ gse_term5(bonds,17);
    energy += /* delta * */ gse_term6(bonds,54);
    energy += /* delta * */ gse_term7(bonds,55);
    energy += /* delta * */ gse_term8(bonds,58);
    {
      int indices[7]={0};
      indices[0]=999999;indices[1]=999999;  /* mark end of bond data */
      indices[2]=999999;indices[3]=999999;
      indices[4]=999999;indices[5]=999999;
      collect_gse_term(indices,0.0,fp_gs);
    } 
  }

  fclose(fp_gs);

  /* added for performance analysis, Joerg 2005 */
  if (Control.output > 0 && myid == 0)
  for(int term=1; term<=62; term++)
  for(int cell=1; cell<=UnitCell.number; cell++) {
    printf("DBG: term cell neighbourloops[2] = %2d %d %d %d\n",
       term, cell,
       Hamiltonian[1][term].k_to_m_number[cell],
       Hamiltonian[2][term].k_to_m_number[cell]);
  }

  /* Joerg: where is CCMcoeffs set? check it! */
  for (j=0, i=0; i<MAX_NUM_CONFIGURATIONS; i++)
    if (CCMcoeffs[i]) { 
       j++;
       printf(" CCMcoeffs[%d]= %f\n", i, CCMcoeffs[i]);
    }
  printf("# nonzero CCMcoeffs: %d\n", j);

  /* This is very slow (only for bond2!) (Joerg 2005, ToDo) */

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    int maxlen=0; // only for user information and debugging
    n_out=0;  // reset for each bond    

    for(i=1;i<=NumberConfigurations;i++) { // total number of clusters

      if (((i%npe)==myid) || PARALLEL_FLAG==0) {

	length = findClusterFromDescriptor(spinConfigurations[i].descriptor,
		   			   configuration);        
	
        if (length > maxlen) maxlen = length;

        // we want to output not to much because not every user knows
        //    what he does ;)
	if (Control.output > 0 && myid == 0
         && ((i-1)%(NumberConfigurations/MAX_out+1) == 0 || i-1<MAX_out)) {
          if (maxlen == length) printf("Bond # = %d : Cluster # = %5d"
                                       " : length = %3d\n", bonds, i, length);
          else                  printf("Bond # = %d : Cluster # = %5d"
            " : length = %3d : maxlen = %3d\n", bonds, i, length, maxlen);
          fflush(stdout); // no buffering (possible time estimations)
          // Joerg: its already printed out on define.C 
	  // printOutDescriptor(spinConfigurations[i].descriptor);
	  // printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
          if (i==MAX_out)
            printf("obtain: output is reduced after %d terms\n",MAX_out);
          maxlen=0;
        }

	
	// s^z s^z : s all s
	
	lhs += term2(i,length,configuration,bonds,2);
	lhs += term3(i,length,configuration,bonds,3);
	lhs += term4(i,length,configuration,bonds,4);
	lhs += term5(i,length,configuration,bonds,5);
	
	// s^z s^+ all s
	
	rhs += term6(i,length,configuration,bonds,6);
	rhs += term7(i,length,configuration,bonds,7);
      
	// s^z s^- all s
	
	rhs += term8(i,length,configuration,bonds,8);
	rhs += term9(i,length,configuration,bonds,9);
	rhs += term10(i,length,configuration,bonds,10);
	rhs += term11(i,length,configuration,bonds,11);
	rhs += term12(i,length,configuration,bonds,12);
	rhs += term13(i,length,configuration,bonds,13);
	
	// s^+ s^z all s
	
	rhs += term14(i,length,configuration,bonds,14);
	rhs += term15(i,length,configuration,bonds,15);
      
      // s^- s^z all s
      
	rhs += term16(i,length,configuration,bonds,16);
	rhs += term17(i,length,configuration,bonds,17);
	rhs += term18(i,length,configuration,bonds,18);
	rhs += term19(i,length,configuration,bonds,19);
	rhs += term20(i,length,configuration,bonds,20);
	rhs += term21(i,length,configuration,bonds,21);

      // s^+ s^- all s

	rhs += term22(i,length,configuration,bonds,22);
	rhs += term23(i,length,configuration,bonds,23);

      // s^- s^+ all s

	rhs += term24(i,length,configuration,bonds,24);
	rhs += term25(i,length,configuration,bonds,25);

      
      // s^+ s^+ all s

	rhs += term26(i,length,configuration,bonds,26);
      
      // s^- s^- all s

	rhs += term27(i,length,configuration,bonds,27);
	rhs += term28(i,length,configuration,bonds,28);
	rhs += term29(i,length,configuration,bonds,29);
	rhs += term30(i,length,configuration,bonds,30); /* slowest 75% square */
	rhs += term31(i,length,configuration,bonds,31); /* t30..34 90% of slowness */
	rhs += term32(i,length,configuration,bonds,32);
	rhs += term33(i,length,configuration,bonds,33);
	rhs += term34(i,length,configuration,bonds,34); /* slowest 31% chain */
	rhs += term35(i,length,configuration,bonds,35);

      // s^- s^- s > 1/2
      
	rhs += term36(i,length,configuration,bonds,36);
	rhs += term37(i,length,configuration,bonds,37);
	rhs += term38(i,length,configuration,bonds,38);
	rhs += term39(i,length,configuration,bonds,39);
	rhs += term40(i,length,configuration,bonds,40);
	rhs += term41(i,length,configuration,bonds,41);
	rhs += term42(i,length,configuration,bonds,42);
	rhs += term43(i,length,configuration,bonds,43);
	rhs += term44(i,length,configuration,bonds,44);
	rhs += term45(i,length,configuration,bonds,45);
      
      // s^z s^- s > 1/2

	rhs += term46(i,length,configuration,bonds,46);
	rhs += term47(i,length,configuration,bonds,47);
	rhs += term48(i,length,configuration,bonds,48);

      // s^- s^z s > 1/2

	rhs += term49(i,length,configuration,bonds,49);
	rhs += term50(i,length,configuration,bonds,50);
	rhs += term51(i,length,configuration,bonds,51);

      // s^+ s^- s > 1/2
     
	rhs += term52(i,length,configuration,bonds,52);

      // s^- s^+ s > 1/2
     
	rhs += term53(i,length,configuration,bonds,53);

      // s^z all s

	rhs += term54(i,length,configuration,bonds,54);
     
      // s^- all s

	rhs += term55(i,length,configuration,bonds,55);
	rhs += term56(i,length,configuration,bonds,56);
	rhs += term57(i,length,configuration,bonds,57);

      // (s^z)^2

	rhs += term58(i,length,configuration,bonds,58);
	rhs += term59(i,length,configuration,bonds,59);
	rhs += term60(i,length,configuration,bonds,60);
	rhs += term61(i,length,configuration,bonds,61);

      // s^+

	rhs += term62(i,length,configuration,bonds,62);
      }
    }
    if (Control.file) {
      int indices[7]={0};
      indices[0]=999999;indices[1]=999999;
      indices[2]=999999;indices[3]=999999;
      indices[4]=999999;indices[5]=999999;
      collect_term(indices,0.0,fp);
    }
    if (myid==0)
      printf("=======================================================\n\n");
  }
  
  if (myid==0)
    printf("\nOBTAIN CCM EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");

  fclose(fp);

}

// s^z s^z : all s 

double term2(int config_index, int length, int **string, int bonds, int term)
{
  
  /* 
     TERM 2: 
     
     F(m) . S(+)(m) . S(z)(k)

     Checked 3-7-2000.
     
  */
  
  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  multiplier=1.0;

  // for(i=1;i<length;i++)      
  // multiplier=multiplier*i; 
  // multiplier=multiplier*length;

  // (length-1)!            -- permutation factor because we sum over m.
  // length                 -- factor from F()

#if DEBUG
  // why it was inside k-loop? Joerg moved it outside Sep05
  findUniqueDescriptorOfCluster(length,configuration,descriptor);
  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
  if (descriptorIndex != config_index) {
    printf("Term 2 Error: descriptorIndex (%d) != config_index (%d)...\n"
           "Exiting!!\n\n", descriptorIndex, config_index);
    exit(1);
  }
#endif
  
  for(k=1;k<=length;k++) {
    
    index = configuration[k][4]; // for every site ...
    
    for(neighbour=1;
        neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	neighbour++) {
      
      /* m ---> k */
      
      new_index = Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
      
      // note that no *CCMcoeffs[config_index] to find in sum
      // because it's the lhs.
      
      sum+=(-0.5*UnitCell.s2[new_index])*
	Hamiltonian[bonds][term].coefficient[index][neighbour];
      
      // -.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=2;
	indices[2]=config_index;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  Hamiltonian[bonds][term].coefficient[index][neighbour];
        collect_term(indices,dummy,fp);
      }
      
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term3(int config_index, int length, int **string, int bonds, int term)
{
  
  /* 

     TERM 3: 
     
     F(k) . S(+)(k) . S(z)(m)

  */
  
  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
//  int descriptorIndex; // not used, just a check
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  multiplier=1.0;

  // for(i=1;i<length;i++) 
  // multiplier=multiplier*i;
  // multiplier=multiplier*length; 

  // (length-1)!            -- permutation factor because we sum over k.
  // length                 -- factor from F()

#if DEBUG
  // moved outside k-loop by Joerg Sep05
  findUniqueDescriptorOfCluster(length,configuration,descriptor);
  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
  
  if (descriptorIndex != config_index) {
    printf("Term 2 Error: descriptorIndex (%d) != config_index (%d)...\n"
           "Exiting!!\n\n", descriptorIndex, config_index);
    exit(1);
  }
#endif

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];      

    for(neighbour=1;
        neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
        neighbour++) {
      
      /* k ---> m */
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      // note that no *CCMcoeffs[config_index] to find in sum
      // because it's the lhs.
      
      sum+=(-0.5*UnitCell.s2[new_index])*
	Hamiltonian[bonds][term].coefficient[index][neighbour];
      
      // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=3;
	indices[2]=config_index;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  Hamiltonian[bonds][term].coefficient[index][neighbour];
        collect_term(indices,dummy,fp);
      }
      
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term4(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 4: 

     H(k,m) . S(+)(k) . S(+)(m).

  */

  int conf[4][6];
  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  double sum;

  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  multiplier=1.0;
  
  // for(i=1;i<=(length-2);i++)
  // multiplier=multiplier*i; 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!       -- permutation factor because we sum over k1 and k2.
  // length*(length-1) -- factor from F()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // k ---> m
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  nn[1]=conf[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(conf[2][1] == nn[1] &&
	     conf[2][2] == nn[2] &&
	     conf[2][3] == nn[3] &&
	     conf[2][4] == nn[4])
	    are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	  
	}
	
	// conf[1] and conf[2] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  // a check here...
	  
#if DEBUG
	  findUniqueDescriptorOfCluster(length,configuration,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex != config_index) {
      printf("Term 4 Error: descriptorIndex (%d) != config_index (%d)...\n"
             "Exiting!!\n\n", descriptorIndex, config_index);
	    exit(1);
	  }
#endif	  
	  // note that no *CCMcoeffs[config_index] to find in sum
	  // because it's the lhs.
	  
	  sum+=
	    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
            indices[1]=4;
	    indices[2]=config_index;
            indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term5(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 5: 

     F(k) . F(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; 
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // k ---> m
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  // ToDo: Why term = bonds here? Bug? Joerg Sep05
          // no diff for LS s=1 SUB6-4 (has no effect?)
          // original was buggy 2nd index was bonds instead of term,
          //  but same results got (LC+square 6-6) tested
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];

	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	  
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=1;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    // now find the multiplying cofficient.
	    
	    multiplier=1.0;
	    
	    // for(i=1;i<conf1_number_of_sites;i++)
	    //   multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over k.
	    // length      -- factor from F(k).
	    
	    // for(i=1;i<conf2_number_of_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m).
	    
	    // rhs this time!
	    
	    // if descriptorIndex is 0, it means no matching cluster
	    // from the `fundamental' set could be found and so 
	    // this case is igmored -- i.e., not saved to disk.
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=CCMcoeffs[descriptorIndex1]
		  *CCMcoeffs[descriptorIndex2]
  		  *Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=5;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
  dummy=Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
                collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term6(int config_index, int length, int **string, int bonds, int term)
{

  
  /* TERM 6: 
     
     S(+)(m) . S(z)(k)
     
  */
  
  int neighbour;
  double multiplier;
  int index,j;
  double sum;
  int new_index;
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  sum=0.0;
  
  if (length != 1)
    return 0.0;
  
  index=string[1][4];
        
  for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
      neighbour++) {
    
    // nearest neighbour is at x,y,z, cell put information
    // into configuration and exit.
    
    /* m ---> k */
    
    new_index=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
    
    // cout << "new Index = " << new_index << endl;
    
    multiplier=1.0;
    multiplier=multiplier*(-0.5*UnitCell.s2[new_index]);
    
    sum+=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=6;
      indices[2]=0; // !Special Case! No CCM ket-state coeffs here!
      indices[3]=0; // This is checked for in solve.C
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
       collect_term(indices,dummy,fp);
    }
    
  }
  return sum;
}

// s^z s^ + all s

double term7(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 7: 
     
     F(k,l) . S(+)(k) . S(+)(m).
     
  */
  
  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,6);
  conf=imatrix(0,MAX_LSUBM+1,0,6);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  // for(i=1;i<length-1;i++)
  //   multiplier=multiplier*i;
  // multiplier=multiplier*(length-1);
  
  // configuration 1
  // (length-2)!     -- permutation factor because we sum over k1 and k2.
  // (length-1)      -- factor from F(). Note that s_k2^+ `captures' one
  //                    of the spin flips but does not contribute to F().
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf[1][4];      

	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(tempConfig[1] == nn[1] &&
	     tempConfig[2] == nn[2] &&
	     tempConfig[3] == nn[3] &&
	     tempConfig[4] == nn[4])
	    are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	}
	
	// conf[1][] and tempConfig[] need to be nearest-neighbours 
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 7: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  findUniqueDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=7;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
              collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term8(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 8: 
     
     -2 * F(m)(l) . S(z)(k) . S(z)(m)
     
     Note that this term generally cancels with term 17.......

     */

  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  int minX;
  int minY;
  int minZ;

  int maxX;
  int maxY;
  int maxZ;

  int order;
  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  int nn[5];

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+4,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1]))
      maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2]))
      maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3]))
      maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;
    minY=0;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=0;
    maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;
    minY=radius;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;
    minY=radius;
    minZ=radius;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
      " not recognised...\nExiting||\n\n", Params.dim);
    exit(1);
  }

  multiplier=1.0;
  //  for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+1);

  // length!               -- permutation factor
  // (length+1)            -- factor from F(m)

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  order=length+1;	            // this is the `m'    
	  
	  configuration[order][1]=x;
	  configuration[order][2]=y;
	  configuration[order][3]=z;
	  configuration[order][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `m'.
	  
	  for(l=1;l<=length;l++)
	    if (configuration[order][1]==configuration[l][1] &&
		configuration[order][2]==configuration[l][2] &&
		configuration[order][3]==configuration[l][3] &&
		configuration[order][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[order][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	      
	      // find the s_k^z, and so use
	      
	      /* m --> k */
	      
	      nn[1]=configuration[order][1]
		+Hamiltonian[bonds][term].m_to_k[index][j][0];
	      nn[2]=configuration[order][2]
		+Hamiltonian[bonds][term].m_to_k[index][j][1];
	      nn[3]=configuration[order][3]
		+Hamiltonian[bonds][term].m_to_k[index][j][2];
	      nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	      
	      temp_multiplier=-2*0.5*UnitCell.s2[configuration[order][4]]*
		0.5*UnitCell.s2[nn[4]];
	      
	      findUniqueDescriptorOfCluster(order,
					    configuration,
					    descriptor);
	      
	      descriptorIndex=findDescriptorIndex(descriptor,
						  NumberConfigurations);
	      
	      if (descriptorIndex > 0)
		sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
		  Hamiltonian[bonds][term].coefficient[index][j];
	      
	      if (Control.file) {
		if (descriptorIndex > 0){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=8;
		  indices[2]=descriptorIndex;
		  indices[3]=0;
		  indices[4]=0;
		  indices[5]=0;
		  dummy=temp_multiplier*
		    Hamiltonian[bonds][term].coefficient[index][j];
                  collect_term(indices,dummy,fp);
		}
	      }
	      
	    }
	  }
	}
      }
    }    
  }
  return sum;  
}

double term9(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 9:
     
     -2 * H(k)(m)(l) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from H()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
	
      /* k ---> m */
	
      nn[1]=configuration[k][1]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
      nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
      numberFlips=1;
	
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
	
      maxNumberFlips=UnitCell.s2[nn[4]];
	
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
	
      if (numberFlips <= maxNumberFlips) {
	  
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4];
	  
	findUniqueDescriptorOfCluster((length+1),configuration,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	  
	// -2 * -0.5*UnitCell.s2[] from coefficient factors in the term 8

	temp_multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	if (descriptorIndex > 0)
	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];;	
	  
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=9;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
	  
      }
    }
  }
  
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term10(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 10: 
     
     - F(m) . F(m) . S(+)(m) . S(z)(k)

     */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m) 1st one
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // F(m) 2nd one
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=
	    Hamiltonian[bonds][term].m_to_k_number[index];neighbour++) {
	
	// find s_k^z and so use
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	
	findUniqueDescriptorOfCluster(conf1_number_of_sites,
				      conf1,
				      descriptor);
	descriptorIndex1=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	findUniqueDescriptorOfCluster(conf2_number_of_sites,
				      conf2,
				      descriptor);
	
	descriptorIndex2=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	multiplier=1.0;
	// for(i=1;i<=conf1_number_of_sites-1;i++)
	//   multiplier=multiplier*i; 
	// multiplier=multiplier*conf1_number_of_sites;
	
	// F(m)
	// (length-1)!       -- permutation factor from F(m)
	// length            -- internal factor in F(m)
	
	// for(i=1;i<=conf2_number_of_sites-1;i++)
	//  multiplier=multiplier*i;
	// multiplier=multiplier*conf2_number_of_sites;
	
	// F(m)
	// (length-1)!       -- permutation factor from F(m)
	// length            -- internal factor in F(m)
	
	multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	  sum+=multiplier
	    *CCMcoeffs[descriptorIndex1]
	    *CCMcoeffs[descriptorIndex2]
	    *Hamiltonian[bonds][term].coefficient[index][neighbour];
	
	if (Control.file) {
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=10;
	    indices[2]=descriptorIndex1;
	    indices[3]=descriptorIndex2;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}

double term11(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 11: 
     
     -2 * H(k)(m) . F(m) . S(+)(k) . S(+)(m)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // H(k,m) <-- k here
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	    
	  /* k ---> m */
	    
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	  
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m];
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=2;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// H(k,m)
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		// F(m)
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    // for(i=1;i<=(conf1_number_of_sites-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites*
	    //  (conf1_number_of_sites-1);
	    
	    // H(k,m)             
	    // (length-2)!        -- because we sum over k and m
	    // length*(length-1)  -- factor due to H().
	    
	    // for(i=1;i<=conf2_number_of_sites-1;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites;
	    
	    // F(m)             
	    // (length-1)!        -- because we sum over k and m
	    // length             -- factor due to F().
	    
	    multiplier=-2.0;
	    
	    // coeff from \tilde s^z \tilde s^-
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]
		*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=11;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term12(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 12:
     
     -2 * F(k) . F(m) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(k)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k)
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // F(m)
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_number_of_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(conf1_number_of_sites,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<=conf1_number_of_sites-1;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_number_of_sites;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<=temp_number_conf2_sites-1;i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites; 
	  
	  // F(m)
	  // (length-1)!       -- permutation factor from F(m)
	  // length            -- internal factor in F(m) because `m' 
	  //                      is `virtual'. (Note that summation over 
	  //                      n.n. gives m thus length ->length+1.)
	  
	  multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coeff from \tilde s^- \tilde s^z
	  
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    sum+=multiplier
	      *CCMcoeffs[descriptorIndex1]
	      *CCMcoeffs[descriptorIndex2]*
	      Hamiltonian[bonds][term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=12;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  


double term13(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 13: 
     
     - F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int **conf1;
  int **conf2;
  int **conf3;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	  
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
          // Joerg: I dont know, why splitting here
          //        but it makes program slow for counter > 10 ???
          //   trimary_max[i] = 3^i, 30*ln3=47.55*ln2 48bit=6*8bit
          // ToDo: simplify spin[i]=000 001 002 010 011 012 020 021 ...
	  if (counter > 30) {
	    printf("TERM 13: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_number_of_sites=1;
		conf2_number_of_sites=1;
		conf3_number_of_sites=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // F(k)
		    conf1_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_number_of_sites][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // F(m)
		    conf2_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_number_of_sites][m]=configuration[temp][m];
		  } else {                   // F(m)
		    conf3_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_number_of_sites][m]=configuration[temp][m];
		  }
		}
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1, descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2, descriptor);
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		findUniqueDescriptorOfCluster(conf3_number_of_sites,
					      conf3, descriptor);	      
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		multiplier=1.0;
		
		// for(i=1;i<=(conf1_number_of_sites-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf1_number_of_sites;
		
		// for(i=1;i<=(conf2_number_of_sites-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_number_of_sites;
		
		// for(i=1;i<=(conf3_number_of_sites-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf3_number_of_sites;
		
		// in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		multiplier=-1.0;
		
		// coeff from \tilde s^z \tilde s^-
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		    && (descriptorIndex3 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]
		    *CCMcoeffs[descriptorIndex3]
		    *Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=13;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

// s^+ s^z 

double term14(int config_index, int length, int **string, int bonds, int term)
{
  
  
  /* TERM 14: 
     
     S(+)(k) . S(z)(m)
     
  */
  
  int neighbour;
  double multiplier;
  int index,j;
  int new_index;
  double sum;

  sum=0.0;

  if (length != 1)
    return 0.0;
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  index=string[1][4];
  
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* k ---> m */
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];

      multiplier=1.0;
      multiplier=multiplier*(-1*0.5*UnitCell.s2[new_index]);
      
      sum+=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
	indices[1]=14;
	indices[2]=0;
	indices[3]=0;
	indices[4]=0;
	indices[5]=0;
	dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
      }
      
  }
  return sum;
}

double term15(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 15: 
     
     F(m) . S(+)(m) . S(+)(k).
     
  */


  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,6);
  conf=imatrix(0,MAX_LSUBM+1,0,6);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  //  for(i=1;i<length-1;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length-1);
  
  // configuration 1
  // (length-2)!     -- permutation factor because we sum over k1 and k2.
  // (length-1)      -- factor from F(). Note that s_k2^+ `captures' one
  //                    of the spin flips but does not contribute to F().
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }    // F(m) 
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; } // s(k)^+
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=conf[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=conf[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=           Hamiltonian[bonds][term].m_to_k[index][j][3];
	  
	    if(tempConfig[1] == nn[1] &&
	       tempConfig[2] == nn[2] &&
	       tempConfig[3] == nn[3] &&
	       tempConfig[4] == nn[4])
	      are_they_neighbours=j;      
	  
	  if (are_they_neighbours) break;
	}
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 15: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  findUniqueDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=15;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+1,0,5);
  return sum;
}

// s^- s^z


double term16(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 16:
     
     -2 * H(m)(k)(l) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  multiplier=1.0;
  //   for(i=1;i<length;i++)
  //  multiplier=multiplier*i; 
  //   multiplier=multiplier*(length+1)*length; 
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length -> length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* m ---> k */
      
      nn[1]=configuration[k][1]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
      nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];

      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4]; // H(m,k)
	
	findUniqueDescriptorOfCluster(length+1,configuration,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=multiplier*2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
	if (descriptorIndex > 0)
	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=16;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
	
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term17(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 17: 
     
     -2 * F(m)(l) * S(z)(k) * S(z)(m)

     */

  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  int minX;
  int minY;
  int minZ;

  int maxX;
  int maxY;
  int maxZ;

  int order;
  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;
  int nn[5];

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+4,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1]))
      maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2]))
      maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3]))
      maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;
    minY=0;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=0;
    maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;
    minY=radius;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;
    minY=radius;
    minZ=radius;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d"
      ") not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

  multiplier=1.0;
  // for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+1);

  // length!               -- permutation factor
  // (length+1)            -- factor from F(k)

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {

	  order=length+1;	            // this is the `k'    
	  
	  configuration[order][1]=x;
	  configuration[order][2]=y;
	  configuration[order][3]=z;
	  configuration[order][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.

	  for(l=1;l<=length;l++)
	    if (configuration[order][1]==configuration[l][1] &&
		configuration[order][2]==configuration[l][2] &&
		configuration[order][3]==configuration[l][3] &&
		configuration[order][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[order][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	      
	      // find the s_k^m, and so use
	      
	      /* k --> m */
	      
	      nn[1]=configuration[order][1]
		+Hamiltonian[bonds][term].k_to_m[index][j][0];
	      nn[2]=configuration[order][2]
		+Hamiltonian[bonds][term].k_to_m[index][j][1];
	      nn[3]=configuration[order][3]
		+Hamiltonian[bonds][term].k_to_m[index][j][2];
	      nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	      
	      temp_multiplier=-2*0.5*UnitCell.s2[configuration[order][4]]*
		0.5*UnitCell.s2[nn[4]];
	      
	      findUniqueDescriptorOfCluster(order,
					    configuration,
					    descriptor);
	      
	      descriptorIndex=findDescriptorIndex(descriptor,
						  NumberConfigurations);
	      
	      if (descriptorIndex > 0)
		sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
		  Hamiltonian[bonds][term].coefficient[index][j];
	      
	      if (Control.file) {
		if (descriptorIndex > 0){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=17;
		  indices[2]=descriptorIndex;
		  indices[3]=0;
		  indices[4]=0;
		  indices[5]=0;
		  dummy=temp_multiplier*
		    Hamiltonian[bonds][term].coefficient[index][j];
            collect_term(indices,dummy,fp);
		}
	      }
	      
	    }
	  }
	}
      }
    }
    
  }
  return sum;  
}

double term18(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 18:
     
     -2 * F(m) . F(k) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    counter--;

    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m)
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // F(k)
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_number_of_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(conf1_number_of_sites,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<conf1_number_of_sites;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_number_of_sites;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<temp_number_conf2_sites;i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites; 
	  
	  // F(m)
	  // (length-1)!       -- permutation factor from F(m)
	  // length            -- internal factor in F(m) because `m' 
	  //                      is `virtual'. (Note that summation over 
	  //                      n.n. gives m thus length ->length+1.)
	  
	  multiplier=2*multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coeff from \tilde s^z \tilde s^-
	  
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    sum+=multiplier
	      *CCMcoeffs[descriptorIndex1]
	      *CCMcoeffs[descriptorIndex2]*
	      Hamiltonian[bonds][term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=18;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term19(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 19: 
     
     -2 * H(k)(m) . F(k) . S(+)(k) . S(+)(m)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // H(m,k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;
      
	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 

	if (are_they_neighbours) {

	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m];
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=2;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// H
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		// H
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    // for(i=1;i<=(conf1_number_of_sites-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites*
	    //  (conf1_number_of_sites-1);
	    
	    // H(m,k)             
	    // (length-2)!        -- because we sum over k and m
	    // length*(length-1)  -- factor due to H().
	    
	    // for(i=1;i<=conf2_number_of_sites-1;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites;
	    
	    // F(m)             
	    // (length-1)!        -- because we sum over k and m
	    // length             -- factor due to F().
	    
	    multiplier=-2.0;
	    
	    // from \tilde s^- \tilde s^z
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=19;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term20(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 20: 
     
     - F(m) . F(m) . S(+)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(k)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(k)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k) 1st one
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // F(k) 2nd one
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=
	    Hamiltonian[bonds][term].k_to_m_number[index];neighbour++) {
	
	// find s_m^z and so use
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
	findUniqueDescriptorOfCluster(conf1_number_of_sites,
				      conf1,
				      descriptor);
	descriptorIndex1=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	findUniqueDescriptorOfCluster(conf2_number_of_sites,
				      conf2,
				      descriptor);
	
	descriptorIndex2=findDescriptorIndex(descriptor,
					     NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	multiplier=1.0;
	// for(i=1;i<=conf1_number_of_sites-1;i++)
	//  multiplier=multiplier*i; 
	// multiplier=multiplier*conf1_number_of_sites;
	
	// F(k)
	// (length-1)!       -- permutation factor from F(k)
	// length            -- internal factor in F(k)
	
	// for(i=1;i<=conf2_number_of_sites-1;i++)
	//  multiplier=multiplier*i;
	// multiplier=multiplier*conf2_number_of_sites;
	
	// F(k)
	// (length-1)!       -- permutation factor from F(k)
	// length            -- internal factor in F(k)
	
	multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^- \tilde s^z
	
	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	  sum+=multiplier
	    *CCMcoeffs[descriptorIndex1]
	    *CCMcoeffs[descriptorIndex2]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];
	
	if (Control.file) {
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=20;
	    indices[2]=descriptorIndex1;
	    indices[3]=descriptorIndex2;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum; 
}


double term21(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 21: 
     
     - F(k) . F(k) . F(m) . S(+)(k) . S(+)(m)
   
    */

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int **conf3;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(k)

	for(m=1;m<5;m++) 
	  conf3[1][m]=configuration[k2][m]; // F(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 21: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_number_of_sites=1;
		conf2_number_of_sites=1;
		conf3_number_of_sites=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {           // F(m)
		    conf1_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_number_of_sites][m]=configuration[temp][m];
		  } else if (spin[j]==1) {    // F(k)
		    conf2_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_number_of_sites][m]=configuration[temp][m];
		  } else {                    // F(k)
		    conf3_number_of_sites++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_number_of_sites][m]=configuration[temp][m];
		  }
		}	      
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
		
                if (descriptorIndex2 <= 0) continue;
		findUniqueDescriptorOfCluster(conf3_number_of_sites,
					      conf3,
					      descriptor);
		
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		multiplier=1.0;
		
		// for(i=1;i<=(conf1_number_of_sites-1);i++)
		//     multiplier=multiplier*i;
		// multiplier=multiplier*conf1_number_of_sites;
		
		// for(i=1;i<=(conf2_number_of_sites-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_number_of_sites;
		
		// for(i=1;i<=(conf3_number_of_sites-1);i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf3_number_of_sites;
		
		  // in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		multiplier=-1.0;
		
		  // from \tilde s^- \tilde s^z
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		    && (descriptorIndex3 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]
		    *CCMcoeffs[descriptorIndex3]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)
		      &&(descriptorIndex3 > 0)){
		      int indices[7]={0};
		      double dummy;
		      indices[0]=config_index;
		      indices[1]=21;
		      indices[2]=descriptorIndex1;
		      indices[3]=descriptorIndex2;
		      indices[4]=descriptorIndex3;
		      indices[5]=0;
		      dummy=multiplier*
			Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }

  
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
 
double term22(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 22:
     
     - 2 * F(m) . S(+)(k) . S(z)(m)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int counter;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i;    // permutation factor
  // multiplier=multiplier*length; // factor from H()
  
  // (length-1)!        -- permutation factor as we sum over k
  // length             -- internal factor from F(). 

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      } 
    }
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      counter=length-1;
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* k ---> m */
      
      nn[1]=configuration[k][1]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
      nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 22: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	findUniqueDescriptorOfCluster(counter,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^+ \tilde s^-
	
	if (descriptorIndex > 0)
	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=22;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=temp_multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
	
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term23(int config_index, int length, int **string, int bonds, int term)
{
  /* TERM 23: 

     - F(m) . F(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(m)
	
	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(k)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	    
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=1;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
	    
                if (descriptorIndex2 <= 0) continue;
	    // for(i=1;i<conf1_number_of_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m)
	    
	    // for(i=1;i<conf2_number_of_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over m.
	    // length      -- factor from F(m)
	    
	    multiplier=-1.0;
	    
	    // coeff from \tilde s^+ \tilde s^-
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=23;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term24(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 24:
     
     -2 * F(k) . S(+)(m) . S(z)(k)
     
  */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int counter;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i;    
  // multiplier=multiplier*length; 
  
  // (length-1)!        -- permutation factor as we sum over m
  // length             -- internal factor from F(). 

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      }
    }
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	neighbour++) {
      
      counter=length-1;
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
      
      /* m ---> k */
      
      nn[1]=configuration[k][1]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
      nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
      
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
      
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 24: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	findUniqueDescriptorOfCluster(counter,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	
	temp_multiplier=2*0.5*UnitCell.s2[nn[4]];
	
	// coeff from \tilde s^+ \tilde s^-
	
	if (descriptorIndex > 0)
	  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];	
	
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=24;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	  }
	}
	
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term25(int config_index, int length, int **string, int bonds, int term)
{
  /* TERM 25: 

     - F(k) . F(k) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // F(k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(k)

	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;
	  
	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	    
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=1;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
	    
                if (descriptorIndex2 <= 0) continue;
	    // multiplier=1.0;
	    
	    // for(i=1;i<conf1_number_of_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites;
	    
	    // configuration 1
	    // (length-1)! -- permutation factor because we sum over k.
	    // length      -- factor from F(k)
	    
	    // for(i=1;i<conf2_number_of_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites;
	    
	    // configuration 2
	    // (length-1)! -- permutation factor because we sum over k
	    // length      -- factor from F(k)
	    
	    multiplier=-1.0;
	    
	    // coeff from \tilde s^+ \tilde s^-
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=25;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term26(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 26: 
     
     S(+)(k) . S(+)(m) 

     */
  
  int tempConfig1[5];
  int tempConfig2[5];
  int nn[5];
  int i,j,k,l,m;
  int are_they_neighbours;
  int index;
  double sum;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  sum=0.0000;

  if(length == 2) {

    for(k=1;k<=2;k++) {
      
      if (k==1) {
	
	for(m=1;m<5;m++) 
	  tempConfig1[m]=string[1][m]; 
	
	for(m=1;m<5;m++) 
	  tempConfig2[m]=string[2][m]; 
	
      } else {

	for(m=1;m<5;m++) 
	  tempConfig1[m]=string[2][m]; 
	
	for(m=1;m<5;m++) 
	  tempConfig2[m]=string[1][m]; 
      }	

      // loop over nearest neighbours
    
      are_they_neighbours=0;
      index=tempConfig1[4];      
	
      for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	
	// nearest neighbour is at x,y,z, temp put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=tempConfig1[1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	nn[2]=tempConfig1[2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	nn[3]=tempConfig1[3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	
	if(tempConfig2[1] == nn[1] &&
	   tempConfig2[2] == nn[2] &&
	   tempConfig2[3] == nn[3] &&
	   tempConfig2[4] == nn[4])
	  are_they_neighbours=j;     
	if (are_they_neighbours) break;
      }
      
      if (are_they_neighbours)
	sum+=1.0*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
      
      if (are_they_neighbours && Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
	indices[1]=26;
	indices[2]=0;
	indices[3]=0;
	indices[4]=0;
	indices[5]=0;
	dummy=1.0*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
      }
      
    }
  }
  return sum;
}

double term27(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 27: 
     
     4 * H(k)(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */

  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int minX;
  int minY;
  int minZ;

  int maxX;
  int maxY;
  int maxZ;


  int order;
  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+4,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1]))
      maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2]))
      maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3]))
      maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radius;
    minY=0;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=0;
    maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;
    minY=radius;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;
    minY=radius;
    minZ=radius;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

  multiplier=1.0;

  //  for(i=1;i<=length;i++)
  //  multiplier=multiplier*i;
  // multiplier=multiplier*(length+2)*(length+1);

  // length!               -- permutation factor
  // (length+2)*(length+1) -- factor from H(k,m)

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {

	  order=length+1;	    
	  
	  configuration[order][1]=x;
	  configuration[order][2]=y;
	  configuration[order][3]=z;
	  configuration[order][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  for(l=1;l<=length;l++)
	    if (configuration[order][1]==configuration[l][1] &&
		configuration[order][2]==configuration[l][2] &&
		configuration[order][3]==configuration[l][3] &&
		configuration[order][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[order][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	      

	      // nearest neighbour is at x,y,z, temp put information
	      // into configuration and exit.
	      
	      /* k --> m */
	      
	      configuration[order+1][1]=configuration[order][1]
		+Hamiltonian[bonds][term].k_to_m[index][j][0];
	      configuration[order+1][2]=configuration[order][2]
		+Hamiltonian[bonds][term].k_to_m[index][j][1];
	      configuration[order+1][3]=configuration[order][3]
		+Hamiltonian[bonds][term].k_to_m[index][j][2];
	      configuration[order+1][4]=
		Hamiltonian[bonds][term].k_to_m[index][j][3];
	      
	      numberFlips=1;
	      
	      for(l=1;l<=order;l++)
		if (configuration[order+1][1]==configuration[l][1] &&
		    configuration[order+1][2]==configuration[l][2] &&
		    configuration[order+1][3]==configuration[l][3] &&
		    configuration[order+1][4]==configuration[l][4])
		  numberFlips++;
	      
	      maxNumberFlips=UnitCell.s2[configuration[order+1][4]];
	      
	      // the number of spin flips that you can have at a given
	      // lattice site is depedent on the quantum spin number
	      // of the site in the UNIT CELL index `length'
	      
	      if (numberFlips <= maxNumberFlips) {
		
		temp_multiplier=4*
		  0.5*UnitCell.s2[configuration[order][4]]*
		  0.5*UnitCell.s2[configuration[order+1][4]];
		
		findUniqueDescriptorOfCluster((order+1),
					      configuration,
					      descriptor);
		
		descriptorIndex=findDescriptorIndex(descriptor,
						    NumberConfigurations);
		
		if (descriptorIndex > 0)
		  sum+=temp_multiplier*CCMcoeffs[descriptorIndex]
		    *Hamiltonian[bonds][term].coefficient[index][j];
		
		if (Control.file) {
		  if (descriptorIndex > 0){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=27;
		    indices[2]=descriptorIndex;
		    indices[3]=0;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=temp_multiplier*
		      Hamiltonian[bonds][term].coefficient[index][j];
            collect_term(indices,dummy,fp);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  return sum;
}

double term28(int config_index, int length, int **string, int bonds, int term)
{
  /* TERM 28: 

     2 * H(k)(m) . H(k)(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // (k)

	for(m=1;m<5;m++) 
	  conf1[2][m]=configuration[k2][m]; // (k)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // (m)
	
	for(m=1;m<5;m++) 
	  conf2[2][m]=configuration[k2][m]; // (m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(conf1[2][1] == nn[1] &&
	     conf1[2][2] == nn[2] &&
	     conf1[2][3] == nn[3] &&
	     conf1[2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=2;
	    conf2_number_of_sites=2;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    
	    // for(i=1;i<=(conf1_number_of_sites-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf1_number_of_sites*
	    // (conf1_number_of_sites-1);
	    
	    // configuration 1
	    // (length-2)!            -- permutation factor because 
	    //                           we sum over k and m.
	    // length*(length-1)      -- factor from H(k,m).
	    
	    // for(i=1;i<=(conf2_number_of_sites-2);i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*conf2_number_of_sites*
	    //  (conf2_number_of_sites-1);
	      
	    // configuration 2
	    // (length-2)!            -- permutation factor because 
	    //                           we sum over k and m.
	    // length*(length-1)      -- factor from H(k,m).
	    
	    multiplier=2.0;
	    
	    // from \tilde s^- \tilde s^-
	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=28;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term29(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 29:
     
     4 * H(k)(m) . F(m) . S(+)(m) . S(z)(k)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  }

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(m)
      conf2[1][m]=configuration[k][m]; // H(m,k) <-- m here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m)
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // H(k,m)
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
 
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf2[1][1]+Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	nn[2]=conf2[1][2]+Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	nn[3]=conf2[1][3]+Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_number_of_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(conf1_number_of_sites,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;

	  // for(i=1;i<conf1_number_of_sites;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_number_of_sites;
	  
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  //  for(i=1;i<=(temp_number_conf2_sites-2);i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites*
	  //  (temp_number_conf2_sites-1); 
	  
	  // H(k,m)
	  // (length-2)!       -- permutation factor from H(k,m)
	  // length*(length-1) -- internal factor in H(k,m) 
	  
	  multiplier=(-4)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- k
	  
	  // coefficient from \tilde s^- \tilde s^-
	  
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    sum+=multiplier
	      *CCMcoeffs[descriptorIndex1]
	      *CCMcoeffs[descriptorIndex2]*
	      Hamiltonian[bonds][term].coefficient[index][neighbour];
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=29;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

/* this is the slowest term, 75% of CPU (Okt2004, Joerg) */
double term30(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 30: 
     
     4 * F(k) . F(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */


  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;

  int maxX, maxY, maxZ;
  int minX, minY, minZ;

  int order;
  int x,y,z,cell;
  int maxPointX, maxPointY, maxPointZ;
  int index;

  int numberFlips, maxNumberFlips;

  int **configuration;

  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int counter;
  int spin[MAX_LSUBM+2];
  int neighbour;
  int temp_x;
  int temp_y;
  int temp_z;
  int temp_cell;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int descriptorIndex1;
  int descriptorIndex2;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) { /* Joerg: change to minR[3] */
    minX=radius;           minY=0;                minZ=0;
    maxX=radius+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;           minY=radius;           minZ=0;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;           minY=radius;           minZ=radius;
    maxX=radius+maxPointX; maxY=radius+maxPointY; maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d) not recognised...\n"
           "Exiting||\n\n",Params.dim);
    exit(1);
  }

  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
          // ToDo: skip x+y+z outside triangle range [0..m]?

	  index=cell;

	  counter=length;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);   /* convert l to binary string spin[] */

            // spin[j]: 0 = F(m)  1 = F(k)
            // copy sites for 0-bit to conf2 and sites for 1-bit to conf1
	    conf1_number_of_sites=0;
	    conf2_number_of_sites=0;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_number_of_sites++;
		for (m=1;m<5;m++) 
		  conf1[conf1_number_of_sites][m]=configuration[j][m];
	      } else {
		conf2_number_of_sites++;
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[j][m];
	      }
	    }
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf1_number_of_sites;m++)
	      if (x==conf1[m][1] &&
		  y==conf1[m][2] &&
		  z==conf1[m][3] &&
		  cell==conf1[m][4])
		numberFlips++;
	    
	    maxNumberFlips = UnitCell.s2[cell];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips > maxNumberFlips) continue; 

            // Took this outside inner loop (Joerg Sep2005)
            temp_number_conf1_sites = conf1_number_of_sites+1;
            conf1[temp_number_conf1_sites][1]=x;
            conf1[temp_number_conf1_sites][2]=y;
            conf1[temp_number_conf1_sites][3]=z;
            conf1[temp_number_conf1_sites][4]=cell;
            
            findUniqueDescriptorOfCluster(temp_number_conf1_sites,
                                          conf1, descriptor);
            descriptorIndex1=findDescriptorIndex(descriptor,
                                                 NumberConfigurations);
            if (descriptorIndex1 <= 0) continue;
	      
            for(neighbour=1;
                neighbour<=Hamiltonian[bonds][term].k_to_m_number[cell];
                neighbour++) {
              
              // nearest neighbour is at x,y,z, cell put information
              // into configuration and exit.
              
              /* k ---> m */
              
              temp_x=x+Hamiltonian[bonds][term].k_to_m[cell][neighbour][0];
              temp_y=y+Hamiltonian[bonds][term].k_to_m[cell][neighbour][1];
              temp_z=z+Hamiltonian[bonds][term].k_to_m[cell][neighbour][2];
              temp_cell=
                       Hamiltonian[bonds][term].k_to_m[cell][neighbour][3];
              
              numberFlips=1;
              
              for(m=1;m<=conf2_number_of_sites;m++)
                if (temp_x==conf2[m][1] &&
                    temp_y==conf2[m][2] &&
                    temp_z==conf2[m][3] &&
                    temp_cell==conf2[m][4])
                  numberFlips++;
              
              maxNumberFlips=UnitCell.s2[temp_cell];
              
              // the number of spin flips that you can have at a given
              // lattice site is depedent on the quantum spin number
              // of the site in the UNIT CELL index `length'
              
              if (numberFlips <= maxNumberFlips) {
                
                

                multiplier=1.0;
                // for(i=1;i<temp_number_conf1_sites;i++)
                //  multiplier=multiplier*i;
                // multiplier=multiplier*temp_number_conf1_sites;
                
                // (length-1)!           -- permutation factor sum over k
                // length                -- factor from F(k)
                
                temp_number_conf2_sites = conf2_number_of_sites+1;
                conf2[temp_number_conf2_sites][1]=temp_x;
                conf2[temp_number_conf2_sites][2]=temp_y;
                conf2[temp_number_conf2_sites][3]=temp_z;
                conf2[temp_number_conf2_sites][4]=temp_cell;

                // for(i=1;i<temp_number_conf2_sites;i++)
                //  multiplier=multiplier*i;
                // multiplier=multiplier*temp_number_conf2_sites;
                
                // (length-1)!           -- permutation factor sum over m
                // length                -- factor from F(m)
                
                multiplier=4*0.5*UnitCell.s2[cell]*
                             0.5*UnitCell.s2[temp_cell];
                
                // coeff from \tilde s^- \tilde s^- 
                
                findUniqueDescriptorOfCluster(temp_number_conf2_sites,
                                              conf2, descriptor);
                descriptorIndex2=findDescriptorIndex(descriptor,
                                                     NumberConfigurations);
                if (descriptorIndex2 <= 0) continue;

                if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
                  sum+=multiplier*CCMcoeffs[descriptorIndex1]
                                 *CCMcoeffs[descriptorIndex2]*
                    Hamiltonian[bonds][term].coefficient[index][neighbour];
                
                if (Control.file) {
                  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
                    int indices[7]={0};
                    double dummy;
                    indices[0]=config_index;
                    indices[1]=30;
                    indices[2]=descriptorIndex1;
                    indices[3]=descriptorIndex2;
                    indices[4]=0;
                    indices[5]=0;
                    dummy=multiplier*
                      Hamiltonian[bonds][term].coefficient[index][neighbour];
                    collect_term(indices,dummy,fp);
                  }
                }
                
              }
            }
	  }
	}
      }
    }
  }
  return sum;
}

double term31(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 31: 
     
     2 * F(m) . F(m) . F(k) . S(+)(m) . S(z)(k)
   
    */
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int **conf3;
  int i, j, k, l, m;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  int **configuration;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // F(m)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // F(m)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    
    if (counter > 30) {
      printf("TERM 31: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_number_of_sites=1;
	  conf2_number_of_sites=1;
	  conf3_number_of_sites=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // F(m)
	      conf1_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf1[conf1_number_of_sites][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // F(m)
	      conf2_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf2[conf2_number_of_sites][m]=configuration[temp][m];
	    } else {                  // F(k)
	      conf3_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf3[conf3_number_of_sites][m]=configuration[temp][m];
	    }
	  }	      
	  
	  index=conf1[1][4];      

          // moved that outside inner loop, Joerg Sep05, speedup
          findUniqueDescriptorOfCluster(conf1_number_of_sites,
                                        conf1, descriptor);
          descriptorIndex1=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex1 <= 0) continue;

          findUniqueDescriptorOfCluster(conf2_number_of_sites,
                                        conf2, descriptor);
          descriptorIndex2=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex2 <= 0) continue;
	  
	  for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf3_number_of_sites;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips++;
	    
	    maxNumberFlips=UnitCell.s2[nn[4]];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips <= maxNumberFlips) {
	      
	      temp_number_conf3_sites = conf3_number_of_sites+1;
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      findUniqueDescriptorOfCluster(temp_number_conf3_sites,
					    conf3, descriptor);
	      descriptorIndex3=findDescriptorIndex(descriptor,
						   NumberConfigurations);
	      if (descriptorIndex3 <= 0) continue;
	      
              multiplier=1.0;
	      
	      // for(i=1;i<conf1_number_of_sites;i++)
	      // multiplier=multiplier*i;
	      // multiplier=multiplier*conf1_number_of_sites;
	      
	      // F(m)
	      // in each case:
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
	      // for(i=1;i<conf2_number_of_sites;i++)
	      // multiplier=multiplier*i;
	      // multiplier=multiplier*conf2_number_of_sites;
	      
	      // F(m)
	      // in each case:
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
		
	      //	      for(i=1;i<temp_number_conf3_sites;i++)
	      //	multiplier=multiplier*i;
	      // multiplier=multiplier*temp_number_conf3_sites;
	      
	      // F(k)
	      // (length-1)!      -- we sum over nearest-neihbours (m).
	      // length           -- internal factor in F().
	      
	      multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- k
	      
	      // from \tilde s^- \tilde s^-
	      
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		  && (descriptorIndex3 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]
		    *CCMcoeffs[descriptorIndex3]*
		    Hamiltonian[bonds][term].coefficient[index][j];
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=31;
		  indices[2]=descriptorIndex1;
		  indices[3]=descriptorIndex2;
		  indices[4]=descriptorIndex3;
		  indices[5]=0;
		  dummy=multiplier*
		    Hamiltonian[bonds][term].coefficient[index][j];
                  collect_term(indices,dummy,fp);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term32(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 32: 
     
     4 * F(k) . F(m) . H(k)(m) . S(+)(k) . S(+)(m)
   
    */
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int **conf3;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k1][m]; // H(k,m) <--- k
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf3[2][m]=configuration[k2][m]; // H(k,m) <--- m
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
		counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 32: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
          temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
          temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
          temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_number_of_sites=1;
		conf2_number_of_sites=1;
		conf3_number_of_sites=2;
		  
		for (j=1;j<=counter;j++) { 
		  temp=tempArray[j];
		  if (spin[j]==0) {          // F(k)
		    conf1_number_of_sites++;
		    for (m=1;m<5;m++) 
		      conf1[conf1_number_of_sites][m]=configuration[temp][m];
		  } else if (spin[j]==1) {   // F(m)
		    conf2_number_of_sites++;
		    for (m=1;m<5;m++) 
		      conf2[conf2_number_of_sites][m]=configuration[temp][m];
		  } else {                    // F(m)
		    conf3_number_of_sites++;
		    for (m=1;m<5;m++) 
		      conf3[conf3_number_of_sites][m]=configuration[temp][m];
		  }
		}	      
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1, descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex1 <= 0) continue;

		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2, descriptor);
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex2 <= 0) continue;

		findUniqueDescriptorOfCluster(conf3_number_of_sites,
					      conf3, descriptor);
		descriptorIndex3=findDescriptorIndex(descriptor,
						     NumberConfigurations);
                if (descriptorIndex3 <= 0) continue;

		multiplier=1.0;
		
		//		for(i=1;i<conf1_number_of_sites;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf1_number_of_sites;
		
		// for(i=1;i<conf2_number_of_sites;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_number_of_sites;
		
		// in each case above F():
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
		// for(i=1;i<=(conf3_number_of_sites-2);i++)
		//    multiplier=multiplier*i;
		// multiplier=multiplier*conf3_number_of_sites*
		//  (conf3_number_of_sites-1);
		
		// H(k,m)
		// (length-2)!        -- we sum over k1 and k2.
		// length*(length-1)  -- internal factor in H().
		
		multiplier=4.0;
		
		// from \tilde s^- \tilde s^-
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		    && (descriptorIndex3 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]
		    *CCMcoeffs[descriptorIndex3]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		  
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		      (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=32;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,fp);
		  }
		}
		  
	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term33(int config_index, int length, int **string, int bonds, int term)
{
  
  /* TERM 33: 
     
     2 * F(k) . F(k) . F(m) . S(+)(k) . S(z)(m)
     
  */
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int **conf3;
  int i, j, k, l, m;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  int **configuration;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // F(k)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // F(k)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }    
    counter--;
    
    if (counter > 30) {
      printf("TERM 33: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  
	  conf1_number_of_sites=1;
	  conf2_number_of_sites=1;
	  conf3_number_of_sites=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {           // F(k)
	      conf1_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf1[conf1_number_of_sites][m]=configuration[temp][m];
	    } else if (spin[j]==1) {    // F(k)
	      conf2_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf2[conf2_number_of_sites][m]=configuration[temp][m];
	    } else {                    // F(m)
	      conf3_number_of_sites++;
	      for (m=1;m<5;m++) 
		conf3[conf3_number_of_sites][m]=configuration[temp][m];
	    }
	  }	      
      
	  are_they_neighbours=0;
	  index=conf1[1][4];      

          // moved outside inner loop, Joerg Sep2005, speedup
          findUniqueDescriptorOfCluster(conf1_number_of_sites,
                                        conf1, descriptor);
          descriptorIndex1=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex1 <= 0) continue;

          findUniqueDescriptorOfCluster(conf2_number_of_sites,
                                          conf2, descriptor);
          descriptorIndex2=findDescriptorIndex(descriptor,
                                               NumberConfigurations);
          if (descriptorIndex2 <= 0) continue;

	  for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	  
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf3_number_of_sites;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips++;
	  
	    maxNumberFlips=UnitCell.s2[nn[4]];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	  
	    if (numberFlips <= maxNumberFlips) {
	      
	      temp_number_conf3_sites=conf3_number_of_sites+1;
	      
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      findUniqueDescriptorOfCluster(temp_number_conf3_sites,
					    conf3, descriptor);
	      descriptorIndex3=findDescriptorIndex(descriptor,
						   NumberConfigurations);
              if (descriptorIndex3 <= 0) continue;

	      multiplier=1.0;
	      
	      // for(i=1;i<conf1_number_of_sites;i++)
	      // multiplier=multiplier*i;
	      // multiplier=multiplier*conf1_number_of_sites;
	      
	      // F(k)
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
	      // for(i=1;i<conf2_number_of_sites;i++)
	      //	multiplier=multiplier*i;
	      // multiplier=multiplier*conf2_number_of_sites;
	    
	      // F(k)
	      // (length-1)!      -- we sum over k.
	      // length           -- internal factor in F().
	      
	      // for(i=1;i<temp_number_conf3_sites;i++)
	      //	multiplier=multiplier*i;
	      // multiplier=multiplier*temp_number_conf3_sites;
	      
	      //F(m)
	      // (length-1)!      -- we sum over nearest-neihbours (m).
	      // length           -- internal factor in F().
	      
	      multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- m
	      
	      // from \tilde s^- \tilde s^-
	      
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)
		  && (descriptorIndex3 > 0))
		sum+=multiplier
		  *CCMcoeffs[descriptorIndex1]
		  *CCMcoeffs[descriptorIndex2]
		  *CCMcoeffs[descriptorIndex3]*
		  Hamiltonian[bonds][term].coefficient[index][j];
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=33;
		  indices[2]=descriptorIndex1;
		  indices[3]=descriptorIndex2;
		  indices[4]=descriptorIndex3;
		  indices[5]=0;
		  dummy=multiplier*
		    Hamiltonian[bonds][term].coefficient[index][j];
                  collect_term(indices,dummy,fp);
		}
	      }
	      
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

/* slowest 31% for chain-6-6 Joerg Sep05 */ 
double term34(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 34:
     
     F(k) . F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf[4]; // Joerg: replaces old **conf1..**conf4 
  int i, j, k, l, m, n, p, mm;
  int temp1, temp2, temp3;
  int k1, k2, ks;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex[4]; // Joerg: replace old descriptor1..4
  double multiplier;
  int conf_number_of_sites[4]; // Joerg: replaces conf[1..4]_number_of_sites
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf[0]=imatrix(0,MAX_LSUBM+1,0,6);   // default: MAX_LSUB=20
  conf[1]=imatrix(0,MAX_LSUBM+1,0,6);
  conf[2]=imatrix(0,MAX_LSUBM+1,0,6);
  conf[3]=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++)
   for(l=1;l<5;l++) {
      conf[0][k][l]=0;
      conf[1][k][l]=0;
      conf[2][k][l]=0;
      conf[3][k][l]=0;
  } 
 
  for(k1=1;k1<=length;k1++)
   for(k2=1;k2<=length;k2++)
    if (k1 < k2) { // old: if (k2 != k1)  Sep05, use 12 <-> 34 symmetry
	
	for(m=1;m<5;m++) {
	  conf[0][1][m]=configuration[k1][m]; // F(k) 12 and 34 are symmetric!
	  conf[1][1][m]=configuration[k1][m]; // F(k)
	  conf[2][1][m]=configuration[k2][m]; // F(m)
	  conf[3][1][m]=configuration[k2][m]; // F(m)
        }
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf[0][1][4];      
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.

	  /* k ---> m */
	  nn[1]= Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]= Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]= Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]= Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if (conf[2][1][1] == conf[0][1][1]+nn[1]
	   && conf[2][1][2] == conf[0][1][2]+nn[2]
	   && conf[2][1][3] == conf[0][1][3]+nn[3]
	   && conf[2][1][4] == nn[4]) { are_they_neighbours=j; break; }
	}
	
	// conf[0][1] and conf[2][1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {     // k3
	    if (l != k1 && l != k2) {  // different from k1 and k2
	      tempArray[counter]=l;    // store all other sites
	      counter++;
	    }
	  }
	  
	  counter--;  // <= length-2 (site could be n-fold appear in config)
	  
	  if (counter > 30) {
	    printf("TERM 34: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
          // Joerg ToDo: why 10? why splitting? (to have no int overflow)
          //         4^10 = 2^20, 4^30 fits to 64bit! dont need splitting

	  for(l=0;l<quartary_max[temp1];l++) {    // 4^counter
	    for(m=0;m<quartary_max[temp2];m++) {
	      for(n=0;n<quartary_max[temp3];n++) {
		
		quartary(l,m,n,counter,spin);  // decode into quartary code
                // sample for counter=3
                // spin[]= 000 001 002 003 010 011 012 013 020 021 ...

                // split the rest of configuration to 4 sub-confs conf[0..3]
		conf_number_of_sites[0]=1; // first site is k1 or k2
		conf_number_of_sites[1]=1;
		conf_number_of_sites[2]=1;
		conf_number_of_sites[3]=1;
		
                // joerg ToDo: simplify code here!
                //  spin[p]:  0     1     2     3
                //  F:       F(k)  F(k)  F(m)  F(m)
		for (p=1;p<=counter;p++) { 
		  temp=tempArray[p];
                  ks = spin[p];
		  conf_number_of_sites[ks]++;
		  for (mm=1;mm<5;mm++)
		    conf[ks][ conf_number_of_sites[ks] ][mm]=
                      configuration[temp][mm];
		}	      
		
                // slowest part (lin.chain)
                for (ks=0; ks<4; ks++) {
  		  findUniqueDescriptorOfCluster(
                    conf_number_of_sites[ks],conf[ks],descriptor);
		  descriptorIndex[ks]=
                    findDescriptorIndex(descriptor,NumberConfigurations);
                  // output here testwise???
                  if (descriptorIndex[ks] <= 0) break; // mostly invalid
                }
                if (ks<4) continue; // skip the rest
                // Joerg: now we have 4 valid indices to configs
                //        its probably faster to test NumConfigs^4 (<=length-1)
                //        if they build our config (length, one same site + miss side)
                // LC-4-4 no term
                // LC-6-6 only Config9: 16*(5,2,2,1)+24*(2,2,1,1)
                // oooooo -> o....o o..o o..o oo + o..o o..o oo oo
                //  sum_len(cfg1..4)=length(conf9)+2 ???
                //  first two configs and last two configs are symmetric!
                //        -> always factor 8 (divide to 12 and 34 and fact=2)

		multiplier=1.0;
		// multiplier = * (conf_number_of_sites[0]-1)!
                //              * (conf_number_of_sites[1]-1)!
                //              * (conf_number_of_sites[2]-1)!
                //              * (conf_number_of_sites[3]-1)!
		// in each case:
		// (length-1)!      -- we sum over k1 and k2.
		// length           -- internal factor in F().
		
                // ToDo: CCMcoeff=0.0 && sum not used? remove? Joerg sep05
		sum+=CCMcoeffs[descriptorIndex[0]]
		    *CCMcoeffs[descriptorIndex[1]]
		    *CCMcoeffs[descriptorIndex[2]]
		    *CCMcoeffs[descriptorIndex[3]]
		  *2*Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		    static int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;  // original config
		    indices[1]=34;
		    indices[2]=descriptorIndex[0]; // modified config1..4
		    indices[3]=descriptorIndex[1];
		    indices[4]=descriptorIndex[2];
		    indices[5]=descriptorIndex[3];
		    dummy= 2   // factor 2 for 3412 sym, Joerg Sep05
		     *Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,fp);
		}
		
	      }
	    }
	  }
	}
  }
  free_imatrix(conf[0],0,MAX_LSUBM+1,0,6);
  free_imatrix(conf[1],0,MAX_LSUBM+1,0,6);
  free_imatrix(conf[2],0,MAX_LSUBM+1,0,6);
  free_imatrix(conf[3],0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;  // not used!
}

double term35(int config_index, int length, int **string, int bonds, int term)
{
  
  /* TERM 35:
     
     4 * H(k)(m) . F(k) . S(+)(k) . S(z)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(k)
      conf2[1][m]=configuration[k][m]; // H(m,k) <-- k here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k)
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // H(k,m)
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf2[1][1]+Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	nn[2]=conf2[1][2]+Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	nn[3]=conf2[1][3]+Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_number_of_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(conf1_number_of_sites,
					conf1,
					descriptor);
	  descriptorIndex1=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	  
                if (descriptorIndex1 <= 0) continue;
	  findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					conf2,
					descriptor);
	  
	  descriptorIndex2=findDescriptorIndex(descriptor,
					       NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	  multiplier=1.0;
	  // for(i=1;i<conf1_number_of_sites;i++)
	  //  multiplier=multiplier*i; 
	  // multiplier=multiplier*conf1_number_of_sites;
	    
	  // F(k)
	  // (length-1)!       -- because we sum over k
	  // length            -- internal factor in F(k)
	  
	  // for(i=1;i<=(temp_number_conf2_sites-2);i++)
	  //  multiplier=multiplier*i;
	  // multiplier=multiplier*temp_number_conf2_sites*
	  //  (temp_number_conf2_sites-1); 
	  
	  // H(k,m)
	  // (length-2)!       -- permutation factor from H(m)
	  // length*(length-1) -- internal factor in H(k,m) 
	  
	  multiplier=(-4)*0.5*UnitCell.s2[nn[4]]; // as nn[] <-- m
	  
	  // coefficient from \tilde s^- \tilde s^-
	    
	  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	    sum+=multiplier
	      *CCMcoeffs[descriptorIndex1]
	      *CCMcoeffs[descriptorIndex2]*
	      Hamiltonian[bonds][term].coefficient[index][neighbour];
	    
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=35;
	      indices[2]=descriptorIndex1;
	      indices[3]=descriptorIndex2;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

// now the ``extra'' terms for s > 1/2.

double term36(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 36:
     
     2 * M(k)(k)(m) . S(+)(k) . S(z)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  //  multiplier=(length+1.0);

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; 

  // multiplier=multiplier*(length+2)*(length+1)*length; 

  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // M(k,k,m)
      
      temp_length=length+1;
      
      // we repeat the `k' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m];

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];

	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 && 
	    numberFlips2 <= maxNumberFlips2)  {
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1]; // `m' is the nearest neighbour.
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=-2*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^- \tilde s^- expansion
	  
	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
		Hamiltonian[bonds][term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {	
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=36;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}
  
double term37(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 37:
     
     2 * M(k)(m)(m) . S(+)(m) . S(z)(z)
     
     NOTE: DIRECTIONALITY TO BE ADDED FOR TRIANGLE!

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+2)*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // M(k,m,m)
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m];
            
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 &&
	    numberFlips2 <= maxNumberFlips2) {	
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1];
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  findUniqueDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
	  
	  temp_multiplier=(-2)*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^- \tilde s^- expansion
	  
	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
		Hamiltonian[bonds][term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=37;
	      indices[2]=descriptorIndex;
              indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}


double term38(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 38: 

     N(k)(k)(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+3,0,5);
  conf=imatrix(0,MAX_LSUBM+3,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  multiplier=1.0;
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*(length+2)*(length+1)*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+2)*(length+1)*length*(length-1) 
  // .................-- factor from N()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `k' and `m' sites for N(k,k,m,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	temp_length=length+2;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k2][m]; // `m'
	
	numberFlips1=1;
	numberFlips2=1;
	
	for(l=1;l<=temp_length-2;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	  if (configuration[k2][1]==conf[l][1] &&
	      configuration[k2][2]==conf[l][2] &&
	      configuration[k2][3]==conf[l][3] &&
	      configuration[k2][4]==conf[l][4])
	    numberFlips2++;
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];
	
	index2=configuration[k2][4];
	maxNumberFlips2=UnitCell.s2[index2];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  nn[1]=configuration[k1][1]+
	    Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (maxNumberFlips2 > 1) &&
	    (numberFlips1 <= maxNumberFlips1) &&
	    (numberFlips2 <= maxNumberFlips2) &&
	    (are_they_neighbours)) {
	  
	  findUniqueDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=38;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+3,0,5);
  return sum;
}

double term39(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 39: 
     
     F(m) . M(k)(k)(m) . S(+)(k) . S(+)(m)
     
    */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  int temp_length;
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {	  

	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // M(k,k,m) k--1st one
	    conf1[2][m]=configuration[k1][m]; // M(k,k,m) k--2nd one
	    conf1[3][m]=configuration[k2][m]; // M(k,k,m) m 
	    conf2[1][m]=configuration[k2][m]; // F(m)     m
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_number_of_sites=3;
	      conf2_number_of_sites=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_number_of_sites][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_number_of_sites][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_number_of_sites;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }

	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/* for(i=1;i<=(conf1_number_of_sites-3);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_number_of_sites*
		  (conf1_number_of_sites-1)*
		  (conf1_number_of_sites-2);
		*/

		// M(k,k,m)             
		// (length-3)!                  -- because we sum over k and m
		// length*(length-1)*(length-2) -- factor due to M().
		
		// for(i=1;i<conf2_number_of_sites;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_number_of_sites;
		
		// F(m)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=2.0;

		// coefficient from \tilde s^- \tilde s^- expansion
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=39;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term40(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 40:
     
     2 * F(k) . H(m)(m) . S(+)(m) . S(z)(k)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // H(m,m) -- 1st m 
	conf1[2][m]=configuration[k][m]; // H(m,m) -- 2nd m
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_number_of_sites=2;
	conf2_number_of_sites=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // F(m)
	    conf1_number_of_sites++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_number_of_sites][m]=configuration[temp][m];
	  } else {
	    // H(k,m)
	    conf2_number_of_sites++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_number_of_sites][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;neighbour<=
	      Hamiltonian[bonds][term].m_to_k_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+
	    Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_number_of_sites;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_number_of_sites;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	  	  
	    temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // H(m,m)
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	    // F(k)
	    findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					  conf2,
					  descriptor);
	
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    /*	    for(i=1;i<=(conf1_number_of_sites-2);i++)
	      multiplier=multiplier*i; 
	    multiplier=multiplier*conf1_number_of_sites*
	      (conf1_number_of_sites-1);
	    */

	    // H(m,m)
	    // (length-2)!       -- because we have m in H(m,m) twice
	    //                      and we sum over it.
	    // length*(length-1) -- internal factor in H(m,m)

	    // for(i=1;i<temp_number_conf2_sites;i++)
	    //  multiplier=multiplier*i;
	    // multiplier=multiplier*temp_number_conf2_sites;

	    // F(k)
	    // (length-1)!       -- permutation + we sum over k
	    // length            -- internal factor in F(k) 
	  
	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]];

	    // factor from 2*S(z)(k)

	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][neighbour];

	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=40;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term41(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 41: 
     
     F(k) . M(m)(m)(k) . S(+)(k) . S(+)(m)
     
    */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  int temp_length;
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // M(m,m,k) m--1st one
	    conf1[2][m]=configuration[k1][m]; // M(m,m,k) m--2nd one
	    conf1[3][m]=configuration[k2][m]; // M(m,m,k) k 
	    conf2[1][m]=configuration[k2][m]; // F(k)     k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_number_of_sites=3;
	      conf2_number_of_sites=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_number_of_sites][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_number_of_sites][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_number_of_sites;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/* for(i=1;i<=(conf1_number_of_sites-3);i++)
		  multiplier=multiplier*i;

		multiplier=multiplier*conf1_number_of_sites*
		  (conf1_number_of_sites-1)*
		  (conf1_number_of_sites-2);
		*/

		// M(m,m,k)             
		// (length-3)!                   -- because we sum over k and m
		// length*(length-1)*(length-2)  -- factor due to M().
		
		//		for(i=1;i<conf2_number_of_sites;i++)
		//  multiplier=multiplier*i;
		// multiplier=multiplier*conf2_number_of_sites;
		
		// F(k)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=2.0; 

		// coefficient from \tilde s^- \tilde s^- expansion
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=41;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term42(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 42:
     
     2 * F(m) . H(k)(k) . S(+)(k) . S(z)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // H(k,k) -- 1st k 
	conf1[2][m]=configuration[k][m]; // H(k,k) -- 2nd k
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_number_of_sites=2;
	conf2_number_of_sites=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // F(m)
	    conf1_number_of_sites++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_number_of_sites][m]=configuration[temp][m];
	  } else {
	    // H(k,m)
	    conf2_number_of_sites++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_number_of_sites][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;
	    neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+
	    Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_number_of_sites;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_number_of_sites;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	  	  
	    temp_number_conf2_sites=conf2_number_of_sites+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // H(k,k)
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex1 <= 0) continue;
	    // F(m)
	    findUniqueDescriptorOfCluster(temp_number_conf2_sites,
					  conf2,
					  descriptor);
	
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
	    multiplier=1.0;
	    /*	    for(i=1;i<=(conf1_number_of_sites-2);i++)
	      multiplier=multiplier*i; 
	    multiplier=multiplier*(conf1_number_of_sites-1)*
	      conf1_number_of_sites;
	    */
	    // H(k,k)
	    // (length-2)!       -- because we have k in H(k,k) twice
	    //                      and we sum over it.
	    // length*(length-1) -- internal factor in H(k,k)

	    /*	    for(i=1;i<temp_number_conf2_sites;i++)
	      multiplier=multiplier*i;
	    multiplier=multiplier*temp_number_conf2_sites;
	    */

	    // F(m)
	    // (length-1)!       -- permutation factor from F(m)
	    // length            -- internal factor in F(m) 
	  
	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]];

	    // factor from 2*S(z)(k)

	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=multiplier
		*CCMcoeffs[descriptorIndex1]
		*CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
	  
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=42;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	      }
	    }

	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term43(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 43: 
     
     H(k)(k) . H(m)(m) . S(+)(k) . S(+)(m)
     
    */
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */


  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips2;
  int maxNumberFlips;
  int numberFlips1;
  int numberFlips2;
  int temp_length;
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	
	if (maxNumberFlips1 > 1 && maxNumberFlips2 > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) k--1st one
	    conf1[2][m]=configuration[k1][m]; // H(k,k) k--2nd one
	    conf2[1][m]=configuration[k2][m]; // H(m,m) m--1st one
	    conf2[2][m]=configuration[k2][m]; // H(m,m) m--2nd one
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_number_of_sites=2;
	      conf2_number_of_sites=2;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_number_of_sites][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_number_of_sites][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_number_of_sites;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];

	      numberFlips2=0;
	      
	      for(m=1;m<=conf2_number_of_sites;m++) {
		if (configuration[k2][1]==conf2[m][1] &&
		    configuration[k2][2]==conf2[m][2] &&
		    configuration[k2][3]==conf2[m][3] &&
		    configuration[k2][4]==conf2[m][4])
		  numberFlips2++;
	      }
	      maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1 &&
		  numberFlips2 <= maxNumberFlips2) {
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;
		/* for(i=1;i<=(conf1_number_of_sites-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_number_of_sites*
		  (conf1_number_of_sites-1);
		*/

		// H(k,k)             
		// (length-2)!        -- because we sum over k and m
		// length*(length-1)  -- factor due to H().
		
		/*		for(i=1;i<=(conf2_number_of_sites-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_number_of_sites*
		  (conf2_number_of_sites-1);
		*/
		// H(m,m)             
		// (length-2)!        -- because we sum over k and m
		// length*(length-1)  -- factor due to H().

		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
		  sum+=CCMcoeffs[descriptorIndex1]
		      *CCMcoeffs[descriptorIndex2]
		      *Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=43;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term44(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 44:
     
     H(k)(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int **conf1, **conf2, **conf3;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int temp_number_conf3_sites;
  int are_they_neighbours;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) <-- 1st k here
	    conf1[2][m]=configuration[k1][m]; // H(k,k) <-- 2nd k here
	    conf2[1][m]=configuration[k2][m]; // F(m)
	    conf3[1][m]=configuration[k2][m]; // F(m)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    if (counter > 30) {
	      printf("TERM 44: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	      
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		    
		  trimary(l,mm,nnn,counter,spin);
		    
		  conf1_number_of_sites=2;
		  conf2_number_of_sites=1;
		  conf3_number_of_sites=1;
		    
		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(k,k)
		      conf1_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_number_of_sites][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(m)
		      conf2_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_number_of_sites][m]=configuration[temp][m];
		    } else {
		      // F(m)
		      conf3_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_number_of_sites][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_number_of_sites;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(k,k)
		    findUniqueDescriptorOfCluster(conf1_number_of_sites,
						  conf1,
						  descriptor);
		    descriptorIndex1=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		    // F(m)
		    findUniqueDescriptorOfCluster(conf2_number_of_sites,
						  conf2,
						  descriptor);
		
		    descriptorIndex2=findDescriptorIndex(descriptor,
							 NumberConfigurations);
	
		
                if (descriptorIndex2 <= 0) continue;
		    // F(m)
		    findUniqueDescriptorOfCluster(conf3_number_of_sites,
						  conf3,
						  descriptor);
		
		    descriptorIndex3=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
		
                if (descriptorIndex3 <= 0) continue;
		    multiplier=1.0;
		    /* for(i=1;i<=conf1_number_of_sites-2;i++)
		      multiplier=multiplier*i; 
		    multiplier=multiplier*(conf1_number_of_sites-1)*
		      conf1_number_of_sites;
		    */
		    // H(k,k)
		    // (length-2)!       -- because we have k in H(k,k) twice
		    //                      and we sum over it.
		    // length(length-1)  -- internal factor in H(k,k)
		    /*
		    for(i=1;i<conf2_number_of_sites;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf2_number_of_sites;
		    */
		    // F(m)
		    // (length-1)!       -- permutation factor from F(m)
		    // length            -- internal factor in F(m) 
		
		    /*		    
                       for(i=1;i<conf3_number_of_sites;i++)
		         multiplier=multiplier*i;
		       multiplier=multiplier*conf3_number_of_sites;
		    */
		    // F(m)
		    // (length-1)!       -- permutation factor from F(m)
		    // length            -- internal factor in F(m) 
		
		    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			(descriptorIndex3 > 0))
		      sum+=CCMcoeffs[descriptorIndex1]
			  *CCMcoeffs[descriptorIndex2]
			  *CCMcoeffs[descriptorIndex3]
			  *Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=44;
			indices[2]=descriptorIndex1;
			indices[3]=descriptorIndex2;
			indices[4]=descriptorIndex3;
			indices[5]=0;
			dummy=
			  Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term45(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 45:
     
     H(m)(m) . F(k) . F(k) . S(+)(k) . S(+)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int **conf1, **conf2, **conf3;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int conf3_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int temp_number_conf3_sites;
  int are_they_neighbours;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  conf3=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 

  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(m,m) <-- 1st m here
	    conf1[2][m]=configuration[k1][m]; // H(m,m) <-- 2nd m here
	    conf2[1][m]=configuration[k2][m]; // F(k)
	    conf3[1][m]=configuration[k2][m]; // F(k)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;

	    if (counter > 30) {
	      printf("TERM 45: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	    
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		  
		  trimary(l,mm,nnn,counter,spin);
		  
		  conf1_number_of_sites=2;
		  conf2_number_of_sites=1;
		  conf3_number_of_sites=1;

		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(m,m)
		      conf1_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_number_of_sites][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(k)
		      conf2_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_number_of_sites][m]=configuration[temp][m];
		    } else {
		      // F(k)
		      conf3_number_of_sites++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_number_of_sites][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_number_of_sites;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(m,m)
		    findUniqueDescriptorOfCluster(conf1_number_of_sites,
						  conf1,
						  descriptor);
		    descriptorIndex1=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		    // F(k)
		    findUniqueDescriptorOfCluster(conf2_number_of_sites,
						  conf2,
						  descriptor);
		
		    descriptorIndex2=findDescriptorIndex(descriptor,
							 NumberConfigurations);
	
                if (descriptorIndex2 <= 0) continue;
		
		    // F(k)
		    findUniqueDescriptorOfCluster(conf3_number_of_sites,
						  conf3,
						  descriptor);
		
		    descriptorIndex3=findDescriptorIndex(descriptor,
							 NumberConfigurations);
		
                if (descriptorIndex3 <= 0) continue;
		
		    multiplier=1.0;
		    /* for(i=1;i<=(conf1_number_of_sites-2);i++)
		      multiplier=multiplier*i; 
		    multiplier=multiplier*(conf1_number_of_sites-1)*
		      conf1_number_of_sites;
		    */
		    // H(m,m)
		    // (length-2)!       -- because we have m in H(m,m) twice
		    //                      and we sum over it.
		    // length(length-1)  -- internal factor in H(m,m)
		    /*
		    for(i=1;i<conf2_number_of_sites;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf2_number_of_sites;
		    */
		    // F(k)
		    // (length-1)!       -- permutation factor from F(k)
		    // length            -- internal factor in F(k) 
		
		    /*		    for(i=1;i<conf3_number_of_sites;i++)
		      multiplier=multiplier*i;
		    multiplier=multiplier*conf3_number_of_sites;
		    */
		    // F(k)
		    // (length-1)!       -- permutation factor from F(k)
		    // length            -- internal factor in F(k) 
		
		    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			(descriptorIndex3 > 0))
		      sum+=CCMcoeffs[descriptorIndex1]
			*CCMcoeffs[descriptorIndex2]
			*CCMcoeffs[descriptorIndex3];

		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=45;
			indices[2]=descriptorIndex1;
			indices[3]=descriptorIndex2;
			indices[4]=descriptorIndex3;
			indices[5]=0;
			dummy=Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf3,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}  

double term46(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 46:
     
     - H(m)(m) . S(+)(m) . S(z)(k)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;
  // for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // H(m,m) <- 1st m
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(m,m) <- 2nd m

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=
	    Hamiltonian[bonds][term].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  Hamiltonian[bonds][term].m_to_k[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].m_to_k[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  findUniqueDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=multiplier*0.5*UnitCell.s2[nn[4]];
	  
	  // coefficient from \tilde s^z \tilde s^- expansion
	  
	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][neighbour];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=46;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}

double term47(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 47: 

     - M(k)(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int **conf;
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+3,0,5);
  conf=imatrix(0,MAX_LSUBM+3,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  multiplier=1.0;
  
  /*  for(i=1;i<=(length-2);i++)
    multiplier=multiplier*i; 
  multiplier=multiplier*(length+1)*length*(length-1);
  */

  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)length(length-1) -- factor from M()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `m' site for M(k,m,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // m -- > k

	  nn[1]=configuration[k1][1]+
	    Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findUniqueDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=-1.0;

	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=47;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+3,0,5);
  return sum;
}

double term48(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 48: 
     
     - F(k) . H(m)(m) . S(+)(k) . S(+)(m)
     
    */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  int temp_length;
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(m,m) m--1st one
	    conf1[2][m]=configuration[k1][m]; // H(m,m) m--2nd one
	    conf2[1][m]=configuration[k2][m]; // F(k)   <- k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].m_to_k[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_number_of_sites=2;
	      conf2_number_of_sites=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_number_of_sites][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_number_of_sites][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_number_of_sites;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;
		/* for(i=1;i<=(conf1_number_of_sites-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_number_of_sites*
		  (conf1_number_of_sites-1);
		*/
		// H(m,m)             
		// (length-2)!        -- because we sum over m
		// length*(length-1)  -- factor due to M().
		/*
		for(i=1;i<conf2_number_of_sites;i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_number_of_sites;
		*/
		// F(k)             
		// (length-1)!        -- because we sum over k
		// length             -- factor due to F().

		multiplier=-1.0; 

		// coefficient from \tilde s^- \tilde s^- expansion
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=48;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term49(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 49: 

     - M(k)(k)(m) . S(+)(k) . S(+)(m).

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int **conf;
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+3,0,5);
  conf=imatrix(0,MAX_LSUBM+3,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  multiplier=1.0;
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*(length+1)*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)*length*(length-1) -- factor from M()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `k' site for M(k,k,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // k -- > m

	  nn[1]=configuration[k1][1]+
	    Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;   

	  if (are_they_neighbours) break; 
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findUniqueDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	 
	  temp_multiplier=-1.0;

	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=49;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+3,0,5);
  return sum;
}

double term50(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 50:
     
     - H(k)(k) . S(+)(k) . S(z)(m)
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  //  for(i=1;i<length;i++)
  //  multiplier=multiplier*i; // permutation factor
  // multiplier=multiplier*(length+1)*length; // factor from M()
  
  // (length-1)!        -- permutation factor as we sum over k
  // (length+1)*length  -- internal factor from H(). Note that 
  //                       summation over n.n. gives m thus 
  //                       length ->length+1.

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; 
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(k,k) <- 2nd k

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
	nn[4]=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  findUniqueDescriptorOfCluster(temp_length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  temp_multiplier=multiplier*0.5*UnitCell.s2[nn[4]]*
	    Hamiltonian[bonds][term].coefficient[index][neighbour];
	  
	  // coefficient from \tilde s^z \tilde s^- expansion
	  
	  if (descriptorIndex > 0)
	    sum+=temp_multiplier*CCMcoeffs[descriptorIndex];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=50;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		Hamiltonian[bonds][term].coefficient[index][neighbour];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}

double term51(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 51: 
     
     - F(m) . H(k)(k) . S(+)(k) . S(+)(m)
     
    */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  int temp_length;
  double sum;
  sum=0.0000;

  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(k,k) k -- 1st one
	    conf1[2][m]=configuration[k1][m]; // H(k,k) k -- 2nd one
	    conf2[1][m]=configuration[k2][m]; // F(m)   m
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+Hamiltonian[bonds][term].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+Hamiltonian[bonds][term].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+Hamiltonian[bonds][term].k_to_m[index][j][2];
	    nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;   

	    if (are_they_neighbours) break;   
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_number_of_sites=2;
	      conf2_number_of_sites=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H(k,k)
		  conf1_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_number_of_sites][m]=configuration[temp][m];
		} else {
		  // F(m)
		  conf2_number_of_sites++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_number_of_sites][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_number_of_sites;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		findUniqueDescriptorOfCluster(conf1_number_of_sites,
					      conf1,
					      descriptor);
		descriptorIndex1=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex1 <= 0) continue;
		findUniqueDescriptorOfCluster(conf2_number_of_sites,
					      conf2,
					      descriptor);
		
		descriptorIndex2=findDescriptorIndex(descriptor,
						     NumberConfigurations);
		
                if (descriptorIndex2 <= 0) continue;
		multiplier=1.0;

		/*		for(i=1;i<=a(conf1_number_of_sites-2);i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf1_number_of_sites*
		  (conf1_number_of_sites-1);
		*/
		// H(k,k)             
		// (length-2)!        -- because we sum over k
		// length*(length-1)  -- factor due to M().
		/*
		for(i=1;i<conf2_number_of_sites;i++)
		  multiplier=multiplier*i;
		multiplier=multiplier*conf2_number_of_sites;
		*/

		// F(m)             
		// (length-1)!        -- because we sum over k and m
		// length             -- factor due to F().

		multiplier=-1.0; 

		// coefficient from \tilde s^- \tilde s^z expansion
		
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
		  sum+=multiplier
		    *CCMcoeffs[descriptorIndex1]
		    *CCMcoeffs[descriptorIndex2]*
		    Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=51;
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term52(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 52: 

     - H(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+3,0,5);
  conf=imatrix(0,MAX_LSUBM+3,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
       
  multiplier=-1.0;
  
  // for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // length(length-1) -- factor from H()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j];
	    counter++;
	  }
	}

	if (counter > length) {
	  printf("Error in term 52...\nExiting!!\n\n");
	  exit(1);
	}

	// a check here...
	
	// now add in extra `m' site for H(m,m)
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];
	
	for(j=1;j<=Hamiltonian[bonds][term].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // m -- > k

	  nn[1]=configuration[k1][1]+
	    Hamiltonian[bonds][term].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    Hamiltonian[bonds][term].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    Hamiltonian[bonds][term].m_to_k[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {	  
	  
	  findUniqueDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=52;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+3,0,5);
  return sum;
}

double term53(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 53: 

     - H(k)(k) . S(+)(k) . S(+)(m).

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+3,0,5);
  conf=imatrix(0,MAX_LSUBM+3,0,5);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  multiplier=-1.0; // note the minus here!!
  
  //  for(i=1;i<=(length-2);i++)
  //  multiplier=multiplier*i; // permutation factor. 
  // multiplier=multiplier*length*(length-1);
  
  // (length-2)!      -- permutation factor because we sum over k1 and k2.
  // (length+1)length(length-1) -- factor from H()

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j];
	    counter++;
	  }
	}
	
	if (counter > length) {
	  printf("Error in term 53...\nExiting!!\n\n");
	  exit(1);
	}

	// a check here...
	
	// now add in extra `k' site for H(k,k)
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];

	for(j=1;j<=Hamiltonian[bonds][term].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // k --> m

	  nn[1]=configuration[k1][1]+
	    Hamiltonian[bonds][term].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    Hamiltonian[bonds][term].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    Hamiltonian[bonds][term].k_to_m[index][j][2];
	  nn[4]=Hamiltonian[bonds][term].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  
	  findUniqueDescriptorOfCluster(temp_length,
					conf,
					descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
              indices[1]=53;
	      indices[2]=descriptorIndex;
              indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=multiplier*
		Hamiltonian[bonds][term].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,fp);
	    }
	  }

	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+3,0,5);
  return sum;
}

// Single-site terms now: s^z

double term54(int config_index, int length, int **string, int bonds, int term)
{
  
  /* 
     TERM 54: 
     
     F(k) . S(+)(k) 

     Wed Jul 10 BST 2002

  */
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  for(k=1;k<=length;k++) {
    
    findUniqueDescriptorOfCluster(length,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    if (descriptorIndex != config_index) {
      printf("Term 54 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex,config_index);
      exit(1);
    }

    index=configuration[k][4];      
    
    sum+=CCMcoeffs[config_index]*
      Hamiltonian[bonds][term].coefficient[index][1];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=54;
      indices[2]=config_index;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=Hamiltonian[bonds][term].coefficient[index][1];
      collect_term(indices,dummy,fp);
    }
  
  }
  
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

// Single-site: s^-

double term55(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 55: 
     
     -2 * F(k) . S(z)(k) 
  
  */
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l, m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;
  
  int minX;
  int minY;
  int minZ;
  
  int maxX;
  int maxY;
  int maxZ;
  
  int order;
  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  int nn[5];
  
  int **configuration;
  
  configuration=imatrix(0,MAX_LSUBM+4,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  int maxPointX;
  int maxPointY;
  int maxPointZ;
  
  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1]))
      maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2]))
      maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3]))
      maxPointZ=abs(string[j][3]);
  }
  
  if (Params.dim == 1) {
    minX=radius;
    minY=0;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=0;
    maxZ=0;
  } else if (Params.dim == 2) {
    minX=radius;
    minY=radius;
    minZ=0;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=0;
  } else if (Params.dim == 3) {
    minX=radius;
    minY=radius;
    minZ=radius;
    maxX=radius+maxPointX;
    maxY=radius+maxPointY;
    maxZ=radius+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d" 
      ") not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }
  
  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  order=length+1;   // this is the `k'    
	  
	  configuration[order][1]=x;
	  configuration[order][2]=y;
	  configuration[order][3]=z;
	  configuration[order][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.
	  
	  for(l=1;l<=length;l++)
	    if (configuration[order][1]==configuration[l][1] &&
		configuration[order][2]==configuration[l][2] &&
		configuration[order][3]==configuration[l][3] &&
		configuration[order][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[order][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    temp_multiplier=0.5*UnitCell.s2[configuration[order][4]]*
	      multiplier;
	    
	    findUniqueDescriptorOfCluster(order,
					  configuration,
					  descriptor);
	    
	    descriptorIndex=findDescriptorIndex(descriptor,
						NumberConfigurations);
	    
	    if (descriptorIndex > 0)
	      sum+=(-2.0)*(-0.5*UnitCell.s2[configuration[order][4]])
		*CCMcoeffs[descriptorIndex]
		*Hamiltonian[bonds][term].coefficient[index][1];
	    
	    if (Control.file) {
	      if (descriptorIndex > 0){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=55;
		indices[2]=descriptorIndex;
		indices[3]=0;
		indices[4]=0;
		indices[5]=0;
		dummy=(-2.0)*(-0.5*UnitCell.s2[configuration[order][4]])*
		  Hamiltonian[bonds][term].coefficient[index][1];
		collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }    
  }
  return sum;  
}

double term56(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 56:
     
     - H(k)(k) . S(+)(k) 
  
  */
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips;
  int numberFlips1;
  int numberFlips2;
  int tempCell;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;
  
  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  conf=imatrix(0,MAX_LSUBM+2,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
  
  multiplier=1.0;
  
  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // H(m,m) <- 1st m
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(m,m) <- 2nd m
      
      index=configuration[k][4];
      
      numberFlips2=0;
      
      for(l=1;l<=temp_length;l++)
	if (configuration[k][1]==conf[l][1] &&
	    configuration[k][2]==conf[l][2] &&
	    configuration[k][3]==conf[l][3] &&
	    configuration[k][4]==conf[l][4])
	  numberFlips2++;
      
      maxNumberFlips2=UnitCell.s2[index];
      
      if (numberFlips2 <= maxNumberFlips2)  {
	
	findUniqueDescriptorOfCluster(temp_length,conf,descriptor);
	descriptorIndex=findDescriptorIndex(descriptor,
					    NumberConfigurations);
	
	if (descriptorIndex > 0)
	  sum+=(-1.0)*CCMcoeffs[descriptorIndex]*
	    Hamiltonian[bonds][term].coefficient[index][1];	
	
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=56;
	    indices[2]=descriptorIndex;
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=(-1.0)*Hamiltonian[bonds][term].coefficient[index][1];
            collect_term(indices,dummy,fp);
	  }
	}
	
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  free_imatrix(conf,0,MAX_LSUBM+2,0,5);
  return sum;
}

double term57(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 57: 
     
     - F(k) . F(k) . S(+)(k) 

  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l,m;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int **configuration;
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int tempCell;
  int **conf1, **conf2;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+2,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
  
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(m)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(m)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_number_of_sites=1;
      conf2_number_of_sites=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m) 1st one
	  conf1_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_number_of_sites][m]=configuration[temp][m];
	} else {
	  // F(m) 2nd one
	  conf2_number_of_sites++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_number_of_sites][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];	

      findUniqueDescriptorOfCluster(conf1_number_of_sites,
				    conf1,
				    descriptor);
      descriptorIndex1=findDescriptorIndex(descriptor,
					   NumberConfigurations);
      
                if (descriptorIndex1 <= 0) continue;
      findUniqueDescriptorOfCluster(conf2_number_of_sites,
				    conf2,
				    descriptor);
      
      descriptorIndex2=findDescriptorIndex(descriptor,
					   NumberConfigurations);
      
                if (descriptorIndex2 <= 0) continue;
      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	sum+=(-1.0)*CCMcoeffs[descriptorIndex1]
	  *CCMcoeffs[descriptorIndex2]
	  *Hamiltonian[bonds][term].coefficient[index][1];
      
      if (Control.file) {
	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	  int indices[7]={0};
	  double dummy;
	  indices[0]=config_index;
	  indices[1]=57;
	  indices[2]=descriptorIndex1;
	  indices[3]=descriptorIndex2;
	  indices[4]=0;
	  indices[5]=0;
	  dummy=(-1.0)*Hamiltonian[bonds][term].coefficient[index][1];
	  collect_term(indices,dummy,fp);
	}
      }
      
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+2,0,5);
  return sum;
}

// s^z s^z : all s 

double term58(int config_index, int length, int **string, int bonds, int term)
{
  
  /* 
     TERM 58: 
     
     2 F(k) . S(+)(k) . S(z)(k)

  */
  
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i,j,k,l;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  double temp_multiplier;
  int **configuration;
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  multiplier=2.0;

  for(k=1;k<=length;k++) {
    
    findUniqueDescriptorOfCluster(length,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    if (descriptorIndex != config_index) {
      printf("Term 58 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex, config_index);
      exit(1);
    }
    
    index=configuration[k][4];      
    
    // note that no *CCMcoeffs[config_index] to find in sum
    // because it's the lhs.
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*
      Hamiltonian[bonds][term].coefficient[index][1];
    
    // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=58;
      indices[2]=config_index;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*(-0.5*UnitCell.s2[index])*
	Hamiltonian[bonds][term].coefficient[index][1];
            collect_term(indices,dummy,fp);
    }
    
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term59(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 59: 
     
     F(k) . S(+)(k) . S(+)(k).
     
  */


  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+1,0,6);
  conf=imatrix(0,MAX_LSUBM+1,0,6);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // F(m)
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(tempConfig[1] == conf[1][1] &&
	   tempConfig[2] == conf[1][2] &&
	   tempConfig[3] == conf[1][3] &&
	   tempConfig[4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 59: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  findUniqueDescriptorOfCluster(length-1,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][1];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=59;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][1];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term60(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 60: 
     
     H_{k,k} . S(+)(k) . S(+)(k).
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */


  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;
  int **conf;
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  configuration=imatrix(0,MAX_LSUBM+1,0,6);
  conf=imatrix(0,MAX_LSUBM+1,0,6);

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // F(m)
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(conf[2][1] == conf[1][1] &&
	   conf[2][2] == conf[1][2] &&
	   conf[2][3] == conf[1][3] &&
	   conf[2][4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=2;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != length) {
	  printf("TERM 60: Error counter != length...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  findUniqueDescriptorOfCluster(length,conf,descriptor);
	  descriptorIndex=findDescriptorIndex(descriptor,
					      NumberConfigurations);
	  
	  if (descriptorIndex > 0)
	    sum+=multiplier*CCMcoeffs[descriptorIndex]*
	      Hamiltonian[bonds][term].coefficient[index][1];	
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=60;
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*Hamiltonian[bonds][term].coefficient[index][1];
            collect_term(indices,dummy,fp);
	    }
	  }
	  
	}
      }
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(conf,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term61(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 61: 

     F(k) . F(k) . S(+)(k) . S(+)(k).
     
  */

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int **conf1;
  int **conf2;
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_the_same_lattice_site;
  int index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_number_of_sites;
  int conf2_number_of_sites;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int **configuration;

  configuration=imatrix(0,MAX_LSUBM+1,0,5);

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=imatrix(0,MAX_LSUBM+1,0,6);
  conf2=imatrix(0,MAX_LSUBM+1,0,6);
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; 
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
	are_they_the_same_lattice_site=0;
	index=conf1[1][4];

	if(conf2[1][1] == conf1[1][1] &&
	   conf2[1][2] == conf1[1][2] &&
	   conf2[1][3] == conf1[1][3] &&
	   conf2[1][4] == conf1[1][4])
	  are_they_the_same_lattice_site=1;      
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_the_same_lattice_site) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_number_of_sites=1;
	    conf2_number_of_sites=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_number_of_sites][m]=configuration[temp][m];
	      } else {
		conf2_number_of_sites++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_number_of_sites][m]=configuration[temp][m];
	      }
	    }	      
	    
	    findUniqueDescriptorOfCluster(conf1_number_of_sites,
					  conf1,
					  descriptor);
	    descriptorIndex1=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex1 <= 0) continue;
	    findUniqueDescriptorOfCluster(conf2_number_of_sites,
					  conf2,
					  descriptor);
	    
	    descriptorIndex2=findDescriptorIndex(descriptor,
						 NumberConfigurations);
	    
                if (descriptorIndex2 <= 0) continue;
	    // now find the multiplying cofficient.
	    	    
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	      sum+=CCMcoeffs[descriptorIndex1]
		  *CCMcoeffs[descriptorIndex2]*
		Hamiltonian[bonds][term].coefficient[index][1];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=61;
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=Hamiltonian[bonds][term].coefficient[index][1];
		collect_term(indices,dummy,fp);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free_imatrix(conf1,0,MAX_LSUBM+1,0,6);
  free_imatrix(conf2,0,MAX_LSUBM+1,0,6);
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double term62(int config_index, int length, int **string, int bonds, int term)
{

  /* TERM 62: 
     
     S(+)(k)

     */
  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  
  int tempConfig1[5];
  int tempConfig2[5];
  int nn[5];
  int i,j,k,l,m;
  int are_they_neighbours;
  int index;
  double sum;

  sum=0.0000;

  if (length == 1) {

    index=string[1][4];

    sum+=Hamiltonian[bonds][term].coefficient[index][1];
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=62;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=Hamiltonian[bonds][term].coefficient[index][1];
      collect_term(indices,dummy,fp); 
    }
      
  }
  return sum;
}

/* Joerg Sep05:
 *  new: collect_gse_term() now collects gse data too
 */
void collect_gse_term(int indices[6],double dummy,FILE *fp_gs){
#if 1
    collect_term(indices,dummy,fp_gs);
#else
    if (Control.output>2 && myid == 0) {
      printf("Cluster Number %d:  GSE TERM %2d",
       indices[0],indices[1]);
      printf(" descriptor_idx (%d,%d,%d,%d) : mult = %f\n",
	     indices[2],indices[3],indices[4],indices[5],dummy);
    }
    if (Control.file) {
      fwrite(indices,sizeof(int),6,fp_gs);
      fwrite(&dummy,sizeof(double),1,fp_gs);
    }
    /* xxzSquare 4-4 maxbuff= 2  terms gs= 8  in= 231  out= 26 */
#endif
}

/* now find terms which contribute to the ground-state energy */

double gse_term1(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 1: 
     
     S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i,j,k,neighbour,index,new_index,**configuration;
  double sum, multiplier;

  sum=0.0;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      multiplier=1.0;

      sum+=multiplier*(-0.5*UnitCell.s2[new_index])*(-0.5*UnitCell.s2[index])*
	Hamiltonian[bonds][term].coefficient[index][neighbour]/UnitCell.number;
      
      // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
      
      {
	int indices[7]={0};
	double dummy;
	indices[0]=0;
        indices[1]=term;
	indices[2]=0;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=multiplier*
	  (-0.5*UnitCell.s2[new_index])*
	  (-0.5*UnitCell.s2[index])*
	  Hamiltonian[bonds][term].coefficient[index][neighbour]
	  /UnitCell.number;
        collect_gse_term(indices,dummy,fp_gs);
      } /* ToDo: collect? */
      
    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
  
  
double gse_term2(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 2: 
     
     4 G_{k,m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, k, l, m,neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;

  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      configuration[1][1]=0;
      configuration[1][2]=0;
      configuration[1][3]=0;
      configuration[1][4]=index;
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      configuration[2][1]=configuration[1][1]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
      configuration[2][2]=configuration[1][2]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
      configuration[2][3]=configuration[1][3]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
      configuration[2][4]=
	Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      findUniqueDescriptorOfCluster(2,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=4.0;

      if (descriptorIndex > 0)
	sum+=multiplier*
	  (-0.5*UnitCell.s2[new_index])*
	  (-0.5*UnitCell.s2[index])
	  *Hamiltonian[bonds][term].coefficient[index][neighbour]
	  *CCMcoeffs[descriptorIndex]
	  /UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    Hamiltonian[bonds][term].coefficient[index][neighbour]
	    /UnitCell.number;
          collect_gse_term(indices,dummy,fp_gs);
	}
      }

    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
 
double gse_term3(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 3: 
     
     4 F_{k} F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, k, l, m,neighbour;
  int index, new_index;
  int descriptor1[MAX_SPIN_LIST];
  int descriptorIndex1;
  int descriptor2[MAX_SPIN_LIST];
  int descriptorIndex2;
  double multiplier;
  int **configuration1;
  int **configuration2;

  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration1=imatrix(0,MAX_LSUBM+1,0,5);
  configuration2=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) {
      configuration1[i][j]=0;
      configuration2[i][j]=0;
    }

  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      configuration1[1][1]=0;
      configuration1[1][2]=0;
      configuration1[1][3]=0;
      configuration1[1][4]=index;
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      configuration2[1][1]=configuration1[1][1]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
      configuration2[1][2]=configuration1[1][2]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
      configuration2[1][3]=configuration1[1][3]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
      configuration2[1][4]=
	Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      findUniqueDescriptorOfCluster(1,configuration1,descriptor1);
      descriptorIndex1=findDescriptorIndex(descriptor1,NumberConfigurations);

                if (descriptorIndex1 <= 0) continue;
      findUniqueDescriptorOfCluster(1,configuration2,descriptor2);
      descriptorIndex2=findDescriptorIndex(descriptor2,NumberConfigurations); 
     
                if (descriptorIndex2 <= 0) continue;
      multiplier=4.0;
      multiplier=multiplier*(-0.5*UnitCell.s2[new_index])*(-0.5*UnitCell.s2[index]);
      
      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
	sum+=multiplier
	  *Hamiltonian[bonds][term].coefficient[index][neighbour]
	  *CCMcoeffs[descriptorIndex1]
	  *CCMcoeffs[descriptorIndex2]/UnitCell.number;
      
      {
	if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
	  indices[1]=term;
	  indices[2]=descriptorIndex1;
	  indices[3]=descriptorIndex2;
	  indices[4]=0;
	  indices[5]=0;
	  dummy=multiplier*
	    Hamiltonian[bonds][term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_gse_term(indices,dummy,fp_gs);
	}
      }
      
    }
  }
  free_imatrix(configuration1,0,MAX_LSUBM+1,0,5);
  free_imatrix(configuration2,0,MAX_LSUBM+1,0,5);
  return sum;
}
  
double gse_term4(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 4: 
     
     -2 F_{m} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, k, l, m,neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;

  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      configuration[1][1]=configuration[1][1]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][0];
      configuration[1][2]=configuration[1][2]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][1];
      configuration[1][3]=configuration[1][3]+
	Hamiltonian[bonds][term].k_to_m[index][neighbour][2];
      configuration[1][4]=
	Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      findUniqueDescriptorOfCluster(1,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=-2.0;

      if (descriptorIndex > 0)
	sum+=multiplier*
	  (-0.5*UnitCell.s2[new_index])*
	  (-0.5*UnitCell.s2[index])
	  *Hamiltonian[bonds][term].coefficient[index][neighbour]
	  *CCMcoeffs[descriptorIndex]/UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    Hamiltonian[bonds][term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_gse_term(indices,dummy,fp_gs);
	}
      }

    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
  

double gse_term5(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 5: 
     
     -2 F_{k} S(z)(k) S(z)(m) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, k, l, m,neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;

  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    for(neighbour=1;neighbour<=Hamiltonian[bonds][term].k_to_m_number[index];
	neighbour++) {
      
      configuration[1][1]=0;
      configuration[1][2]=0;
      configuration[1][3]=0;
      configuration[1][4]=index;

      new_index=Hamiltonian[bonds][term].k_to_m[index][neighbour][3];
      
      findUniqueDescriptorOfCluster(1,configuration,descriptor);
      descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
      
      multiplier=-2.0;

      if (descriptorIndex > 0)
	sum+=multiplier*
	  (-0.5*UnitCell.s2[new_index])*
	  (-0.5*UnitCell.s2[index])
	  *Hamiltonian[bonds][term].coefficient[index][neighbour]
	  *CCMcoeffs[descriptorIndex]
	  /UnitCell.number;
      
      {
	if (descriptorIndex > 0) {
	  int indices[7]={0};
	  double dummy;
	  indices[0]=0;
          indices[1]=term;
	  indices[2]=descriptorIndex;
          indices[3]=0;
	  indices[4]=0;
          indices[5]=0;
	  dummy=multiplier*
	    (-0.5*UnitCell.s2[new_index])*
	    (-0.5*UnitCell.s2[index])*
	    Hamiltonian[bonds][term].coefficient[index][neighbour]
	    /UnitCell.number;
	  collect_gse_term(indices,dummy,fp_gs);
	}
      }

    }
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
  
double gse_term6(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 6: Single Site
     
     S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int i,j,k,neighbour,index,new_index,**configuration;
  double sum, multiplier;

  sum=0.0;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*
      Hamiltonian[bonds][term].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	Hamiltonian[bonds][term].coefficient[index][1]
	/UnitCell.number;
      collect_gse_term(indices,dummy,fp_gs);
    }
      
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
  
double gse_term7(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 7: 
     
     -2 F_{m} S(z)(k) 

     Wed Jul 10 BST 2002
     
  */

  int i, j, k, l, m,neighbour;
  int index, new_index;
  int descriptor[MAX_SPIN_LIST];
  int descriptorIndex;
  double multiplier;
  int **configuration;

  double sum;
  sum=0.0000;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=index;
      
    findUniqueDescriptorOfCluster(1,configuration,descriptor);
    descriptorIndex=findDescriptorIndex(descriptor,NumberConfigurations);
    
    multiplier=-2.0;
    
    if (descriptorIndex > 0)
      sum+=multiplier*
	(-0.5*UnitCell.s2[index])*
	Hamiltonian[bonds][term].coefficient[index][1]
	*CCMcoeffs[descriptorIndex]/UnitCell.number;
    
    {
      if (descriptorIndex > 0) {
	int indices[7]={0};
	double dummy;
	indices[0]=0;
        indices[1]=term;
	indices[2]=descriptorIndex;
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=multiplier*
	  (-0.5*UnitCell.s2[index])*
	  Hamiltonian[bonds][term].coefficient[index][1]
	  /UnitCell.number;
	collect_gse_term(indices,dummy,fp_gs);
      }
    }
    
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}

double gse_term8(int bonds, int term)
{

 /* 
     GROUND-STATE ENERGY TERM 8: Single Site
     
     (S(z)(k))^2 

     Wed Jul 10 BST 2002
     
  */

  int i,j,k,neighbour,index,new_index,**configuration;
  double sum, multiplier;

  sum=0.0;

  if (!Hamiltonian[bonds][term].evaluate) return 0; /* skip that term */

  configuration=imatrix(0,MAX_LSUBM+1,0,5);
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=0;
  
  // get the multiplying coefficient correct first!

  for(index=1;index<=UnitCell.number;index++) {
    
    multiplier=1.0;
    
    sum+=multiplier*(-0.5*UnitCell.s2[index])*(-0.5*UnitCell.s2[index])*
      Hamiltonian[bonds][term].coefficient[index][1]/UnitCell.number;
    
    {
      int indices[7]={0};
      double dummy;
      indices[0]=0;
      indices[1]=term;
      indices[2]=0;
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=multiplier*
	(-0.5*UnitCell.s2[index])*
	(-0.5*UnitCell.s2[index])*
	Hamiltonian[bonds][term].coefficient[index][1]
	/UnitCell.number;
      collect_gse_term(indices,dummy,fp_gs);
    }
      
  }
  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  return sum;
}
  
